-- -------------------------------------------------------------------------- --
-- CharScanner by kunda                                                       --
-- -------------------------------------------------------------------------- --
--                                                                            --
-- CharScanner scans Basic Character Data, Talent Specifications, Equipment,  --
-- Achievements and Boss Kill Statistics from target unit or group and shows  --
-- this data in a movable frame. Any scanned unit data is temporarily saved   --
-- (in memory not in SavedVariables) and can be viewed or deleted any time.   --
--                                                                            --
-- ONE click and ONE look shows the most important character data!            --
--                                                                            --
-- -------------------------------------------------------------------------- --
--                                                                            --
-- Notes:                                                                     --
-- - The scanned unit MUST BE in inspect range!                               --
-- - There is no 'Inspect Glyphs API'! :-(                                    --
-- - Maximum time for one scan is: 1 Second (default).                        --
--    !!! Change this value in the options menu if you have a high latency or --
--        if you get a 'Scan incomplete!' status message if you scan yourself.--                                                      --
-- - GroupScan:                                                               --
--    - Raid roster changes between a 'GroupScan'-Button click and end of     --
--      groupscan are discarded!                                              --
--    - GroupScan will stop if:                                               --
--        All group/raid members has been scanned. or                         --
--        You enter combat! or                                                --
--        You click on: 'STOP SCAN'                                           --
-- - LDB is supported.                                                        --
--    But: LibStub, CallbackHandler-1.0 & LibDataBroker-1.1 is not included.  --
--         This is intended and I will not change this!                       --
-- - CharScanner will never ever save scanned data to SavedVariables.         --
-- - Entering combat and entering/leaving an instance will stop any scan.     --
-- - Memory usage:                                                            --
--    - Base memory usage without a scan: ~500KB                              --
--    - Base memory with populated class talent data: ~700KB                  --
--    - One unit: ~25KB                                                       --
--    - One class talent data: ~20KB                                          --
--    = 25 scanned units (with each class) needs ~1300KB (1.3MB) memory       --
--      there is room for some improvement                                    --
--                                                                            --
-- - slash commands: /cs - /scan - /charscanner                               --
--          example: /scan         : CharScanner GUI                          --
--          example: /scan options : CharScanner Options                      --
--                                                                            --
-- -------------------------------------------------------------------------- --

-- ---------------------------------------------------------------------------------------------------------------------
CharScanner = CreateFrame("Frame")    -- container
CharScanner_Options = {}              -- SavedVariable options table

local _G                           = _G
local type                         = _G.type
local mod                          = _G.mod
local pairs                        = _G.pairs
local select                       = _G.select
local tonumber                     = _G.tonumber
local math_floor                   = _G.math.floor
local string_sub                   = _G.string.sub
local GetTalentTabInfo             = _G.GetTalentTabInfo
local GetTalentInfo                = _G.GetTalentInfo
local GetTalentLink                = _G.GetTalentLink
local GetStatistic                 = _G.GetStatistic
local GetComparisonStatistic       = _G.GetComparisonStatistic
local GetAchievementComparisonInfo = _G.GetAchievementComparisonInfo
local GetAchievementInfo           = _G.GetAchievementInfo
local GetCategoryNumAchievements   = _G.GetCategoryNumAchievements
local UnitGUID                     = _G.UnitGUID
local GetTime                      = _G.GetTime

local L = CharScanner_Locales         -- localization table
local DATA = CharScanner_Data         -- achievement & statistic data table

local mult2 = 10^(2 or 0)             -- x.xx
local mult4 = 10^(4 or 0)             -- x.xxxx

local AchievementIDTable = {}         -- Achievement ID->name table (build after event PLAYER_LOGIN triggered)
local AchievementIDTableCount = 0     -- Achievement ID table count
local StatisticIDTable = {}           -- Statistic ID->name table (build after event PLAYER_LOGIN triggered)
local StatisticIDTableCount = 0       -- Statistic ID table count
local inCombat = false                -- false = out of combat | true = in combat
local inPartyOrRaid = false           -- false = NOT in party or raid | true = in party or raid
local rosterhack = 0                  -- execute RAID_ROSTER_UPDATE block only if GetNumRaidMembers() has changed
local bugfix                          -- -> workaround for achievementFunctions.selectedCategory bug (Blizzard_AchievementUI.lua)
local memusage = false                -- ugly

local SingleScanIsActive = false      -- false = SingleScan is NOT active | true = SingleScan is active
local SingleScanTimerStart = 0        -- Timer (single scan) start (time for one scan)
local DataStep = 0                    -- upvalue for the different information sources (eg. events)
local GroupScanNext = false           -- 
local GroupScanIsActive = false       -- false = GroupScan is NOT active | true = GroupScan is active
local GroupScanTimerStart = 0         -- Timer (group scan) start (time for whole group scan)

local GUIstats = {}                   -- Achievement/Statistic GUI: collapse values
local GUIselectedGUID                 -- current shown char
local GUImaxHistory = 20              -- history scroll frame size
local GUIhistory = {}                 -- history list (scanned units)
local GUIhistoryCount = 0             -- history list (scanned units) count

local TalentTreeFrameIsShown = false  -- for mouseover
local GearFrameIsShown = false        -- for mouseover
local HistoryFrameIsActivated = false --
local NumScannedChars = 0             -- number of scanned characters

local IsUnitInpected = false          -- scanned unit: true: "Inspect" | false: "Character" (it's me)
local SCANGUID                        -- current scan GUID
local SCAN = {}                       -- scanned unit data (key = GUID)

local myMAX_TALENT_TABS = 3           -- Blizzard default is 3 too 
local myMAX_NUM_TALENTS = 44          -- Blizzard default is 40 (how strange) -> we use 44, because: 11 (tier) * 4 (column) = 44 (button)

local StatsBarHeight = 12             -- Achievement/Statistic Bar Height
local FrameWidth = 400-- TODO

local IMAGE_Button = "Interface\\Addons\\CharScanner\\CharScanner-images-button.tga"
local IMAGE_Icon   = "Interface\\Addons\\CharScanner\\CharScanner-images-icons.tga"

local playerRealm = GetCVar("realmName") -- current realm name
local playerGUID = ""                    -- current character GUID (set after event PLAYER_LOGIN triggered)

local DefaultStatsViewData = { -- WotLK default view | customizable by user TODO
	[1] = {lvl = 1, name1 = "WotLK", index1 = 3, name2 = nil,       index2 = nil, name3 = nil, index3 = nil},
	[2] = {lvl = 2, name1 = "WotLK", index1 = 3, name2 = "raid10p", index2 = 2,   name3 = nil, index3 = nil},
	[3] = {lvl = 2, name1 = "WotLK", index1 = 3, name2 = "raid25p", index2 = 3,   name3 = nil, index3 = nil}
}

local color = { -- e.g.: color.White.hex | color.Yellow.rgb[1]...[3]
	DarkGrey    = {hex = "333333", rgb = {0.2,  0.2,  0.2  }},
	Grey        = {hex = "808080", rgb = {0.5,  0.5,  0.5  }},
	Red         = {hex = "ff0000", rgb = {1,    0,    0    }},
	White       = {hex = "ffffff", rgb = {1,    1,    1    }},
	Level1      = {hex = "ffffff", rgb = {1,    1,    1    }},
	Level2      = {hex = "cccccc", rgb = {0.8,  0.8,  0.8  }},
	Level3      = {hex = "999999", rgb = {0.6,  0.6,  0.6  }},
	White4      = {hex = "ffffff", rgb = {1,    1,    1    }},
	LightYellow = {hex = "ffff7f", rgb = {1,    1,    0.49 }},
	Yellow      = {hex = "ffd100", rgb = {1,    0.82, 0    }},
	Green       = {hex = "00bf00", rgb = {0,    0.75, 0    }}
}

local factionimgAlliance = "Interface\\TargetingFrame\\UI-PVP-Alliance"
local factionimgHorde    = "Interface\\TargetingFrame\\UI-PVP-Horde"

local classimg = "Interface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes"
local classes = {-- the classes table is populated with talent data if a class (with lvl>9) is scanned the first time
	DRUID       = {number =  1, color = "ff7d0a", icon = {0.7578125,  0.97265625, 0.01953125, 0.234375}}, -- {194/256,249/256,  5/256, 60/256}},
	HUNTER      = {number =  2, color = "abd473", icon = {0.01953125, 0.234375,   0.26953125, 0.484375}}, -- {  5/256, 60/256, 69/256,124/256}},
	MAGE        = {number =  3, color = "69ccf0", icon = {0.265625,   0.48046875, 0.01953125, 0.234375}}, -- { 68/256,123/256,  5/256, 60/256}},
	PALADIN     = {number =  4, color = "f58cba", icon = {0.01953125, 0.234375,   0.51953125, 0.734375}}, -- {  5/256, 60/256,133/256,188/256}},
	PRIEST      = {number =  5, color = "ffffff", icon = {0.51171875, 0.7265625,  0.26953125, 0.484375}}, -- {131/256,186/256, 69/256,124/256}},
	ROGUE       = {number =  6, color = "fff569", icon = {0.51171875, 0.7265625,  0.01953125, 0.234375}}, -- {131/256,186/256,  5/256, 60/256}},
	SHAMAN      = {number =  7, color = "2459ff", icon = {0.265625,   0.48046875, 0.26953125, 0.484375}}, -- { 68/256,123/256, 69/256,124/256}},
	WARLOCK     = {number =  8, color = "9482c9", icon = {0.7578125,  0.97265625, 0.26953125, 0.484375}}, -- {194/256,249/256, 69/256,124/256}},
	WARRIOR     = {number =  9, color = "c79c6e", icon = {0.01953125, 0.234375,   0.01953125, 0.234375}}, -- {  5/256, 60/256,  5/256, 60/256}},
	DEATHKNIGHT = {number = 10, color = "c41f3b", icon = {0.265625,   0.48046875, 0.51953125, 0.734375}}  -- { 68/256,123/256,133/256,188/256}}
}

local guildRank = {
	[1] = GUILD_RANK0_DESC, -- 0 | Guild Master
	[2] = GUILD_RANK1_DESC, -- 1 | Officer
	[3] = GUILD_RANK2_DESC, -- 2 | Veteran
	[4] = GUILD_RANK3_DESC, -- 3 | Member
	[5] = GUILD_RANK4_DESC  -- 4 | Initiate
}

-- Equipment
local slots = {
	[1] = "HeadSlot",
	[2] = "NeckSlot",
	[3] = "ShoulderSlot",
	[4] = "BackSlot",
	[5] = "ChestSlot",
	[6] = "WristSlot",
	[7] = "HandsSlot",
	[8] = "WaistSlot",
	[9] = "LegsSlot",
	[10] = "FeetSlot",
	[11] = "Finger0Slot",
	[12] = "Finger1Slot",
	[13] = "Trinket0Slot",
	[14] = "Trinket1Slot",
	[15] = "MainHandSlot",
	[16] = "SecondaryHandSlot",
	[17] = "RangedSlot"	
}
local slotsNum = #slots

local slotButton = {
	[1]  = {id =  1, sid =  1, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Head"},
	[2]  = {id =  2, sid =  2, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Neck"},
	[3]  = {id =  3, sid =  3, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Shoulder"},
	[4]  = {id =  4, sid = 15, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Chest"}, -- Back
	[5]  = {id =  5, sid =  5, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Chest"},
	[6]  = {id = -1, sid = -1, IQ = false, texture="Interface\\PaperDollInfoFrame\\UI-GearManager-LeaveItem-Opaque"}, -- Interface\\Paperdoll\\UI-PaperDoll-Slot-Shirt
	[7]  = {id = -1, sid = -1, IQ = false, texture="Interface\\PaperDollInfoFrame\\UI-GearManager-LeaveItem-Opaque"}, -- Interface\\Paperdoll\\UI-PaperDoll-Slot-Tabart
	[8]  = {id =  6, sid =  9, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Wrists"},
	[9]  = {id = 15, sid = 16, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-MainHand"}, -- middle
	[10] = {id = 16, sid = 17, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-SecondaryHand"}, -- middle
	[11] = {id = 17, sid = 18, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Ranged"}, -- middle
	[12] = {id = -1, sid = -1, IQ = false, texture="Interface\\PaperDollInfoFrame\\UI-GearManager-LeaveItem-Opaque"}, -- Interface\\Paperdoll\\UI-PaperDoll-Slot-Ammo" -- middle
	[13] = {id = 14, sid = 14, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Trinket"},
	[14] = {id = 13, sid = 13, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Trinket"},
	[15] = {id = 12, sid = 12, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Finger"},
	[16] = {id = 11, sid = 11, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Finger"},
	[17] = {id = 10, sid =  8, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Feet"},
	[18] = {id =  9, sid =  7, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Legs"},
	[19] = {id =  8, sid =  6, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Waist"},
	[20] = {id =  7, sid = 10, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Hands"}
}

local itemQuality = {
	[1] = {name = ITEM_QUALITY0_DESC, color = "9d9d9d"}, -- 0 | poor
	[2] = {name = ITEM_QUALITY1_DESC, color = "ffffff"}, -- 1 | common
	[3] = {name = ITEM_QUALITY2_DESC, color = "1eff00"}, -- 2 | uncommon
	[4] = {name = ITEM_QUALITY3_DESC, color = "0070dd"}, -- 3 | rare
	[5] = {name = ITEM_QUALITY4_DESC, color = "a335ee"}, -- 4 | epic
	[6] = {name = ITEM_QUALITY5_DESC, color = "ff8000"}, -- 5 | legendary
	[7] = {name = ITEM_QUALITY6_DESC, color = "e6cc80"}, -- 6 | artifact (not in game)
	[8] = {name = ITEM_ACCOUNTBOUND,  color = "e6cc80"}  -- 7 | account bound
}
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
local function Print(...)
	print("|cff"..color.LightYellow.hex.."CharScanner:|r", ...)
end

local function BuildIDTable()
	for i=1, GetCategoryNumAchievements(-1) do
		local id, name = GetAchievementInfo(-1, i)
		if id and name then
			AchievementIDTable[id] = name
			AchievementIDTableCount = AchievementIDTableCount + 1
			--TEST code for future Achievement updates START
			--if CharScanner_Options.AchievementTest == nil then CharScanner_Options.AchievementTest = {} end
			--if not CharScanner_Options.AchievementTest[id] then
			--	CharScanner_Options.AchievementTest[id] = name
			--end
			--TEST code for future Achievement updates END
		end
	end
	--print("AchievementIDTableCount=", AchievementIDTableCount) --TEST code for future Achievement updates
	
	for i=1, GetCategoryNumAchievements(-2) do
		local id, name = GetAchievementInfo(-2, i)
		if id and name then
			StatisticIDTable[id] = name
			StatisticIDTableCount = StatisticIDTableCount + 1
			--TEST code for future Statistic updates START
			--local quantity = GetStatistic(id)
			--if CharScanner_Options.StatisticTest == nil then CharScanner_Options.StatisticTest = {} end
			--if not CharScanner_Options.StatisticTest[id] then
			--	CharScanner_Options.StatisticTest[id] = name
			--end
			--TEST code for future Statistic updates END
		end
	end
	--print("StatisticIDTableCount=", StatisticIDTableCount) --TEST code for future Statistic updates
end

local function CreateScanDataByGUID(guid)
	SCAN[guid] = {}
	SCAN[guid].name = nil                      -- scanned unit: base info - name
	SCAN[guid].level = nil                     -- scanned unit: base info - level
	SCAN[guid].race = nil                      -- scanned unit: base info - race
	SCAN[guid].class = nil                     -- scanned unit: base info - class (localized)
	SCAN[guid].classEN = nil                   -- scanned unit: base info - class (english)
	SCAN[guid].faction = nil                   -- scanned unit: base info - faction (english)
	SCAN[guid].guildName = nil                 -- scanned unit: base info - guild name
	SCAN[guid].guildRankName = nil             -- scanned unit: base info - guild rank name
	SCAN[guid].guildRankIndex = nil            -- scanned unit: base info - guild rank index
	SCAN[guid].talentGroups = nil              -- scanned unit: talent - talent groups (count)
	SCAN[guid].activeTalentGroup = nil         -- scanned unit: talent - active talent group

	SCAN[guid].talents = {}                    -- scanned unit: all spend talent points
	SCAN[guid].talentGroup1name = nil          -- scanned unit: talent - talent group 1 name with most points (eg: Affliction)
	SCAN[guid].talentGroup1pointsString = nil  -- scanned unit: talent - talent group 1 points (eg: 55/11/6)
	SCAN[guid].talentGroup1pointsUnspent = nil -- scanned unit: talent - talent group 1 points unspend (eg: 0)
	SCAN[guid].talentGroup1icon = nil          -- scanned unit: talent - talent group 1 icon
	SCAN[guid].talentGroup2name = nil          -- scanned unit: talent - talent group 2 name with most points (eg: Destruction)
	SCAN[guid].talentGroup2pointsString = nil  -- scanned unit: talent - talent group 2 points (eg: 3/10/58)
	SCAN[guid].talentGroup2pointsUnspent = nil -- scanned unit: talent - talent group 2 points unspend (eg: 1)
	SCAN[guid].talentGroup2icon = nil          -- scanned unit: talent - talent group 2 icon

	SCAN[guid].gear = {}                       -- scanned unit: itemLink from every gear slot
	SCAN[guid].gearStats = {}                  -- scanned unit: basic gear stats from every gear slot
	SCAN[guid].itemlevelAvg = nil              -- scanned unit: itemlevel - average item level & lowest item level & highest item level
	SCAN[guid].itemlevelQuality = nil          -- scanned unit: itemlevel - quality

	SCAN[guid].DATA = {}                       -- scanned unit: achievement & statistic data
	SCAN[guid].statsAchievementTotal = 0       -- scanned unit: summary: achievements - total achievement points
	SCAN[guid].statsAchievementCount = 0       -- scanned unit: summary: achievements - total achievement points
	SCAN[guid].statsStatisticTotal = 0         -- scanned unit: summary: statistics - total achievement points
	SCAN[guid].statsStatisticCount = 0         -- scanned unit: summary: statistics - total achievement points
	SCAN[guid].statsStatisticBossKills = 0     -- scanned unit: summary: statistics - total achievement points
	
	SCAN[guid].time = nil                      -- scanned unit: scan time (eg.: '22:19')
	SCAN[guid].scanTime = nil                  -- scanned unit: time that elapsed for scan (eg.: '0.362 seconds')
end

local function NodeDisplay(button, state)
	if state then
		_G[button]:Show()
	else
		_G[button]:Hide()
	end
	for i = 1, #GUIstats do
		if GUIstats[i].name == button then
			GUIstats[i].collapse = state
			break
		end
	end

	local height = 0
	local visibleNode = {}
	local si = 1
	for i = 1, #GUIstats do
		if GUIstats[i].collapse then
			if GUIstats[i].level == 2 then
				visibleNode[si] = i
				si = si + 1
			elseif GUIstats[i].level == 3 then
				for j = 1, #GUIstats do
					if GUIstats[j].level == 2 and GUIstats[j].index1 == GUIstats[i].index1 and GUIstats[j].collapse then
						visibleNode[si] = i
						si = si + 1
					end
				end
			elseif GUIstats[i].level == 4 then

				for j = 1, #GUIstats do
					if GUIstats[j].level == 2 and GUIstats[j].index1 == GUIstats[i].index1 and GUIstats[j].collapse then--and
						for k = 1, #GUIstats do
							if GUIstats[k].level == 3 and GUIstats[k].index2 == GUIstats[i].index2 and GUIstats[k].collapse then
								visibleNode[si] = i
								si = si + 1
							end
						end
					end
				end

			end
		end
	end
	for i = 1, #visibleNode do
		height = height + GUIstats[visibleNode[i]].height
	end
	table.wipe(visibleNode)
	return height
end

local function IQ_GetItemQuality(unit, who)
	local itemRarityCount = {}
	for i = 1, 8 do
		itemRarityCount[i] = 0
	end
	local quality = ""
	for i = 1, slotsNum do
		local id = GetInventoryItemLink(unit, _G[who..slots[i]]:GetID())
		if id then
			local linkID = select(3, id:find("item:(%d+)"))
			local itemRarity = select(3, GetItemInfo(linkID))
			itemRarityCount[itemRarity+1] = itemRarityCount[itemRarity+1] + 1
		end
	end
	local x = 0
	for i = 1, #itemRarityCount do
		if itemRarityCount[i] > 0 then
			if x == 1 then
				quality = quality..", "
				x = 0
			end
			quality = quality.."|cff"..color.White.hex..itemRarityCount[i].."x|r |cff"..itemQuality[i].color..itemQuality[i].name.."|r"
			x = 1
		end
	end
	return quality
end

local function IQ_GetItemLevel(unit, slot, index, slotPlain)
	local id = GetInventoryItemLink(unit, GetInventorySlotInfo(slotPlain))-- slot:GetID())
	if not SCAN[SCANGUID].gear[index] then SCAN[SCANGUID].gear[index] = {} end

	if id then
		local linkID = select(3, id:find("item:(%d+)"))
		local itemRarity = select(3, GetItemInfo(linkID))
		local iLevel = select(4, GetItemInfo(linkID))

		SCAN[SCANGUID].gear[index].iLink = id or ""
		SCAN[SCANGUID].gear[index].iLvl  = iLevel or 0
		SCAN[SCANGUID].gear[index].iRar  = itemRarity or 0

		for key, value in pairs(GetItemStats(id)) do
			if key and value and type(value) == "number" then
				if not SCAN[SCANGUID].gearStats[key] then
					SCAN[SCANGUID].gearStats[key] = value
				else
					SCAN[SCANGUID].gearStats[key] = SCAN[SCANGUID].gearStats[key] + value
				end
			end
		end

		return iLevel or 0
	else

		SCAN[SCANGUID].gear[index].iLink = ""
		SCAN[SCANGUID].gear[index].iLvl  = 0
		SCAN[SCANGUID].gear[index].iRar  = 0

		return -1
	end
end

local function IQ_2HWeaponCheck(unit, slot1, slot2)
	local id = GetInventoryItemLink(unit, slot1:GetID())
	local twohandweapon = false
	if id then
		local linkID = select(3, id:find("item:(%d+)"))
		local itemEquipLoc = select(9, GetItemInfo(linkID))
		if itemEquipLoc == "INVTYPE_2HWEAPON" then
			local id2 = GetInventoryItemLink(unit, slot2:GetID())
			if not id2 then
				twohandweapon = true
			end
		end
	end
	return twohandweapon
end

local function SlashHandler(msg)
	msg = strtrim(string.lower(msg or ""))
	if msg == "options" then
		InterfaceOptionsFrame_OpenToCategory("CharScanner")
	elseif msg == "mem" then
		memusage = not memusage
		if memusage then
			Print(L["mem_desc1"])
		else
			Print(L["mem_desc2"])
		end		
		CharScanner:UpdateHistoryStatus()
	else
		CharScanner:MainFrameToggle()
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:InitOptions()
	SlashCmdList["CHARSCANNER"] = SlashHandler
	SLASH_CHARSCANNER1 = "/cs"
	SLASH_CHARSCANNER2 = "/scan"
	SLASH_CHARSCANNER3 = "/charscanner"

	-- version = internal options version
	if CharScanner_Options.version and CharScanner_Options.version == 1 then
		CharScanner_Options = {}
		CharScanner_Options.version = 2
		Print("Options reset.")
	end

	if CharScanner_Options.version        == nil then CharScanner_Options.version        = 2     end
	if CharScanner_Options.scanButton     == nil then CharScanner_Options.scanButton     = false end
	if CharScanner_Options.scanButtonLock == nil then CharScanner_Options.scanButtonLock = false end
	if CharScanner_Options.scanButtonSize == nil then CharScanner_Options.scanButtonSize = 24    end
	if CharScanner_Options.chatoutput     == nil then CharScanner_Options.chatoutput     = false end
	if CharScanner_Options.scanDuration   == nil then CharScanner_Options.scanDuration   = 1     end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:LDBcheck()
	if LibStub and LibStub:GetLibrary("CallbackHandler-1.0", true) and LibStub:GetLibrary("LibDataBroker-1.1", true) then
		LibStub:GetLibrary("LibDataBroker-1.1"):NewDataObject("CharScanner", {
			type = "launcher",
			icon = IMAGE_Button,
			iconCoords = {0.078125, 0.921875, 0.078125, 0.921875}, -- (5/64, 59/64, 5/64, 59/64)
			OnClick = function(self, button)
				CharScanner:MainFrameToggle()
			end,
		})
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:CreateInterfaceOptions()
	local options = CreateFrame("FRAME", "CharScanner_InterfaceOptions")
	options.name = "CharScanner"

	local title = options:CreateFontString(nil, "ARTWORK", "GameFontNormalLarge")
	title:SetPoint("TOPLEFT", 16, -16)
	title:SetText("CharScanner")
	title:SetJustifyH("LEFT")
	title:SetJustifyV("TOP")

	local GUI = CreateFrame("Button", "CharScanner_InterfaceOptions_GUI", CharScanner_InterfaceOptions, "UIPanelButtonTemplate")
	CharScanner_InterfaceOptions_GUI:SetPoint("TOPLEFT", title, "BOTTOMLEFT", 0, -10)
	CharScanner_InterfaceOptions_GUI:SetWidth(120)
	CharScanner_InterfaceOptions_GUI:SetHeight(24)
	CharScanner_InterfaceOptions_GUIText:SetText(L["Open GUI"])
	CharScanner_InterfaceOptions_GUI:SetScript("OnClick", function(self)
		InterfaceOptionsFrame_Show()  -- hide InterfaceOptionsFrame
		HideUIPanel(GameMenuFrame)    -- hide GameMenuFrame
		CharScanner:MainFrameToggle() -- show CharScanner GUI
	end)

	local slashcommandText = options:CreateFontString(nil, "ARTWORK", "GameFontHighlightSmall")
	slashcommandText:SetPoint("LEFT", GUI, "RIGHT", 10, 0)
	slashcommandText:SetText("(|cff"..color.LightYellow.hex.."/cs|r - |cff"..color.LightYellow.hex.."/scan|r - |cff"..color.LightYellow.hex.."/charscanner|r)")
	slashcommandText:SetNonSpaceWrap(true)

	local ScanButton = CreateFrame("CheckButton", "CharScanner_InterfaceOptions_ScanButton", CharScanner_InterfaceOptions, "InterfaceOptionsCheckButtonTemplate")
	CharScanner_InterfaceOptions_ScanButton:SetPoint("TOPLEFT", GUI, "BOTTOMLEFT", 0, -10)
	CharScanner_InterfaceOptions_ScanButtonText:SetText(L["Show SCAN-Button"])
	CharScanner_InterfaceOptions_ScanButton:SetChecked(CharScanner_Options.scanButton)
	CharScanner_InterfaceOptions_ScanButton.tooltipText = L["Shows a small movable SCAN-Button if your target is a player."]

	local ScanButtonLock = CreateFrame("CheckButton", "CharScanner_InterfaceOptions_ScanButtonLock", CharScanner_InterfaceOptions, "InterfaceOptionsCheckButtonTemplate")
	CharScanner_InterfaceOptions_ScanButtonLock:SetPoint("TOPLEFT", ScanButton, "BOTTOMLEFT", 10, 0)
	CharScanner_InterfaceOptions_ScanButtonLockText:SetText(L["Lock SCAN-Button"])
	CharScanner_InterfaceOptions_ScanButtonLock:SetChecked(CharScanner_Options.scanButtonLock)

	CharScanner_InterfaceOptions_ScanButton:SetScript("OnClick", function(self)
		CharScanner_Options.scanButton = not CharScanner_Options.scanButton
		if CharScanner_Options.scanButton then
			CharScanner_ScanButtonFrame:Show()
			BlizzardOptionsPanel_CheckButton_Enable(CharScanner_InterfaceOptions_ScanButtonLock, true)
		else
			CharScanner_ScanButtonFrame:Hide()
			BlizzardOptionsPanel_CheckButton_Disable(CharScanner_InterfaceOptions_ScanButtonLock)
		end
	end)
	CharScanner_InterfaceOptions_ScanButtonLock:SetScript("OnClick", function(self)
		CharScanner_Options.scanButtonLock = not CharScanner_Options.scanButtonLock
		if CharScanner_Options.scanButtonLock then
			CharScanner_ScanButtonFrame_Mover:Hide()
		else
			CharScanner_ScanButtonFrame:Show()
			CharScanner_ScanButtonFrame_Mover:Show()
		end
	end)

	if CharScanner_Options.scanButton then
		BlizzardOptionsPanel_CheckButton_Enable(CharScanner_InterfaceOptions_ScanButtonLock, true)
	else
		BlizzardOptionsPanel_CheckButton_Disable(CharScanner_InterfaceOptions_ScanButtonLock)
	end

	local chatOutput = CreateFrame("CheckButton", "CharScanner_InterfaceOptions_chatOutput", CharScanner_InterfaceOptions, "InterfaceOptionsCheckButtonTemplate")
	CharScanner_InterfaceOptions_chatOutput:SetPoint("TOPLEFT", ScanButtonLock, "BOTTOMLEFT", -10, -10)
	CharScanner_InterfaceOptions_chatOutputText:SetText(L["Chat output"])
	CharScanner_InterfaceOptions_chatOutput:SetChecked(CharScanner_Options.chatoutput)
	CharScanner_InterfaceOptions_chatOutput:SetScript("OnClick", function(self)
		CharScanner_Options.chatoutput = not CharScanner_Options.chatoutput
	end)
	CharScanner_InterfaceOptions_chatOutput.tooltipText = L["chat_desc"]

	local scanduration = CreateFrame("Slider", "CharScanner_InterfaceOptions_scanduration", CharScanner_InterfaceOptions, "OptionsSliderTemplate")
	CharScanner_InterfaceOptions_scanduration:SetWidth(200)
	CharScanner_InterfaceOptions_scanduration:SetHeight(16)
	CharScanner_InterfaceOptions_scanduration:SetPoint("TOPLEFT", chatOutput, "BOTTOMLEFT", 4, -30)
	CharScanner_InterfaceOptions_scandurationText:SetPoint("LEFT")
	CharScanner_InterfaceOptions_scandurationText:SetText(L["Maximum time for one scan"]..":")
	CharScanner_InterfaceOptions_scandurationLow:SetText("1")
	CharScanner_InterfaceOptions_scandurationHigh:SetText("2")
	CharScanner_InterfaceOptions_scanduration:SetMinMaxValues(10, 20)
	CharScanner_InterfaceOptions_scanduration:SetValueStep(1)
	CharScanner_InterfaceOptions_scanduration:SetValue(CharScanner_Options.scanDuration*10)
	local valuetext = scanduration:CreateFontString("CharScanner_InterfaceOptions_scanduration_valuetext", "OVERLAY", "GameFontNormal")
	valuetext:SetPoint("TOP", scanduration, "BOTTOM", 0, 0)
	valuetext:SetJustifyH("CENTER")
	valuetext:SetText(CharScanner_Options.scanDuration)
	CharScanner_InterfaceOptions_scanduration:SetScript("OnValueChanged", function(self, value)
		value = value/10
		CharScanner_InterfaceOptions_scanduration_valuetext:SetText(value)
		CharScanner_Options.scanDuration = value
	end)
	CharScanner_InterfaceOptions_scanduration.tooltipText = L["scanduration_desc"]

	InterfaceOptions_AddCategory(options)
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:InitializeFrame()
	-- initial Achievement & Statistics collapse menu appearance
	local opt = DefaultStatsViewData
	for i = 1, #opt do
		CharScanner:NodeClick(opt[i].lvl, opt[i].name1, opt[i].index1, opt[i].name2, opt[i].index2, opt[i].name3, opt[i].index3)
	end
end

function CharScanner:MainFrameToggle()
	if CharScanner_MainFrame:IsShown() then
		CharScanner_MainFrame:Hide()
	else
		CharScanner_MainFrame:Show()
	end
end

function CharScanner:ScanFrameShow()
	CharScanner_MainFrame_Base:Show()
	CharScanner_MainFrame_Talent:Show()
	CharScanner_MainFrame_Equipment:Show()
	CharScanner_MainFrame_Stats:Show()
	CharScanner_MainFrame_Summary:Show()
	CharScanner_MainFrame_Status:Show()
	if HistoryFrameIsActivated then
		CharScanner_HistoryFrame:Show()
	end
	CharScanner_MainFrame_Headline_ButtonCollapseNormal:SetTexture("Interface\\Buttons\\UI-Panel-CollapseButton-Up")
	CharScanner_MainFrame_Headline_ButtonCollapsePush:SetTexture("Interface\\Buttons\\UI-Panel-CollapseButton-Up")
	if NumScannedChars == 0 then
		CharScanner_MainFrame_Headline_ButtonCollapse:Disable()
		CharScanner_MainFrame_Headline_ButtonCollapseBorderColor:SetTexture(0.2, 0.2, 0.2, 1)
		CharScanner_MainFrame_Headline_ButtonHistory:Disable()
		CharScanner_MainFrame_Headline_ButtonHistoryBorderColor:SetTexture(0.2, 0.2, 0.2, 1)
	else
		CharScanner_MainFrame_Headline_ButtonCollapse:Enable()
		CharScanner_MainFrame_Headline_ButtonCollapseBorderColor:SetTexture(0.4, 0.4, 0.4, 1)
		CharScanner_MainFrame_Headline_ButtonHistory:Enable()
		CharScanner_MainFrame_Headline_ButtonHistoryBorderColor:SetTexture(0.4, 0.4, 0.4, 1)
	end
end

function CharScanner:ScanFrameHide()
	CharScanner_MainFrame_Base:Hide()
	CharScanner_MainFrame_Talent:Hide()
	CharScanner_MainFrame_Equipment:Hide()
	CharScanner_MainFrame_Stats:Hide()
	CharScanner_MainFrame_Summary:Hide()
	CharScanner_MainFrame_Status:Hide()
	CharScanner_HistoryFrame:Hide()
	CharScanner_MainFrame_Headline_ButtonHistory:Disable()
	CharScanner_MainFrame_Headline_ButtonHistoryBorderColor:SetTexture(0.2, 0.2, 0.2, 1)
	CharScanner_MainFrame_Headline_ButtonCollapseNormal:SetTexture("Interface\\Buttons\\UI-Panel-ExpandButton-Up")
	CharScanner_MainFrame_Headline_ButtonCollapsePush:SetTexture("Interface\\Buttons\\UI-Panel-ExpandButton-Up")
	if NumScannedChars == 0 then
		CharScanner_MainFrame_Headline_ButtonCollapse:Disable()
		CharScanner_MainFrame_Headline_ButtonCollapseBorderColor:SetTexture(0.2, 0.2, 0.2, 1)
	else
		CharScanner_MainFrame_Headline_ButtonCollapse:Enable()
		CharScanner_MainFrame_Headline_ButtonCollapseBorderColor:SetTexture(0.4, 0.4, 0.4, 1)
	end
end

function CharScanner:ScanFrameToggle(force)
	if force then
		CharScanner:ScanFrameShow()
	else
		if CharScanner_MainFrame_Base:IsShown() then
			CharScanner:ScanFrameHide()
		else
			CharScanner:ScanFrameShow()
		end
	end
end

function CharScanner:CheckStatusFramePosition()
	if CharScanner_MainFrame_Base:IsShown() then
		CharScanner_MainFrame_Status:SetPoint("TOPLEFT", CharScanner_MainFrame_Summary, "BOTTOMLEFT", 0, 0)
	else
		CharScanner_MainFrame_Status:SetPoint("TOPLEFT", CharScanner_MainFrame_Headline, "BOTTOMLEFT", 0, 0)
	end
end

function CharScanner:GroupScanToggle()
	if inPartyOrRaid then
		CharScanner_MainFrame_Headline_ButtonGroupScan:Enable()
		CharScanner_MainFrame_Headline_ButtonGroupScanBorderColor:SetTexture(0.4, 0.4, 0.4, 1)
	else
		CharScanner_MainFrame_Headline_ButtonGroupScan:Disable()
		CharScanner_MainFrame_Headline_ButtonGroupScanBorderColor:SetTexture(0.2, 0.2, 0.2, 1)
	end
end

function CharScanner:HistoryFrameToggle()
	if CharScanner_HistoryFrame:IsShown() then
		CharScanner_HistoryFrame:Hide()
		HistoryFrameIsActivated = false
	else
		CharScanner_HistoryFrame:Show()
		HistoryFrameIsActivated = true
	end
end

function CharScanner:GearFrameToggle()
	if GearFrameIsShown then
		CharScanner_MainFrame_Gear:Hide()
		CharScanner_MainFrame_Stats:SetPoint("TOPLEFT", CharScanner_MainFrame_Equipment, "BOTTOMLEFT", 0, 0)
		CharScanner_MainFrame_Equipment_ButtonCollapseNormal:SetTexture("Interface\\Buttons\\UI-Panel-ExpandButton-Up")
		CharScanner_MainFrame_Equipment_ButtonCollapsePush:SetTexture("Interface\\Buttons\\UI-Panel-ExpandButton-Up")
		GearFrameIsShown = false
	else
		CharScanner_MainFrame_Gear:Show()
		CharScanner_MainFrame_Stats:SetPoint("TOPLEFT", CharScanner_MainFrame_Gear, "BOTTOMLEFT", 0, 0)
		CharScanner_MainFrame_Equipment_ButtonCollapseNormal:SetTexture("Interface\\Buttons\\UI-Panel-CollapseButton-Up")
		CharScanner_MainFrame_Equipment_ButtonCollapsePush:SetTexture("Interface\\Buttons\\UI-Panel-CollapseButton-Up")
		GearFrameIsShown = true
	end
end

function CharScanner:TalentTreeFrameToggle()
	if TalentTreeFrameIsShown then
		CharScanner_MainFrame_TalentTree:Hide()
		CharScanner_MainFrame_Equipment:SetPoint("TOPLEFT", CharScanner_MainFrame_Talent, "BOTTOMLEFT", 0, 0)
		CharScanner_MainFrame_Talent_ButtonCollapseNormal:SetTexture("Interface\\Buttons\\UI-Panel-ExpandButton-Up")
		CharScanner_MainFrame_Talent_ButtonCollapsePush:SetTexture("Interface\\Buttons\\UI-Panel-ExpandButton-Up")
		TalentTreeFrameIsShown = false
	else
		CharScanner_MainFrame_TalentTree:Show()
		CharScanner_MainFrame_Equipment:SetPoint("TOPLEFT", CharScanner_MainFrame_TalentTree, "BOTTOMLEFT", 0, 0)
		CharScanner_MainFrame_Talent_ButtonCollapseNormal:SetTexture("Interface\\Buttons\\UI-Panel-CollapseButton-Up")
		CharScanner_MainFrame_Talent_ButtonCollapsePush:SetTexture("Interface\\Buttons\\UI-Panel-CollapseButton-Up")
		TalentTreeFrameIsShown = true
	end
end

function CharScanner:MainFramePos()
	if CharScanner_Options.MainFrame_relP then
		CharScanner_MainFrame:ClearAllPoints()
		CharScanner_MainFrame:SetPoint(CharScanner_Options.MainFrame_relP, CharScanner_Options.MainFrame_posX, CharScanner_Options.MainFrame_posY)
	else
		CharScanner_MainFrame:ClearAllPoints()
		CharScanner_MainFrame:SetPoint("CENTER", 0, 0)
	end
end

function CharScanner:MainFrameHide()
	PlaySound("igQuestListClose")
	CharScanner_InterfaceOptions_GUI:Enable()
end

function CharScanner:MainFrameShow()
	PlaySound("igQuestListOpen")
	CharScanner_InterfaceOptions_GUI:Disable()
	CharScanner:MainFramePos()
	CharScanner:UpdateHistory()
end

function CharScanner:SaveMainFramePos()
	local point, relativeTo, relativePoint, x, y = CharScanner_MainFrame:GetPoint()
	CharScanner_Options.MainFrame_posX = x
	CharScanner_Options.MainFrame_posY = y
	CharScanner_Options.MainFrame_relP = relativePoint
end

function CharScanner:SaveScanButtonPos()
	local point, relativeTo, relativePoint, x, y = CharScanner_ScanButtonFrame:GetPoint()
	CharScanner_Options.ScanButton_posX = x
	CharScanner_Options.ScanButton_posY = y
	CharScanner_Options.ScanButton_relP = relativePoint
end

function CharScanner:ScanButtonPos()
	if CharScanner_Options.ScanButton_relP then
		CharScanner_ScanButtonFrame:ClearAllPoints()
		CharScanner_ScanButtonFrame:SetPoint(CharScanner_Options.ScanButton_relP, CharScanner_Options.ScanButton_posX, CharScanner_Options.ScanButton_posY)
	else
		CharScanner_ScanButtonFrame:ClearAllPoints()
		CharScanner_ScanButtonFrame:SetPoint("CENTER", 0, 0)
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:CreateMainFrame()
	-- frame/button templates START
	local function TemplateHeadlineButton(button, small)
		local texture = button:CreateTexture(nil, "BORDER")
		texture:SetPoint("TOPLEFT", 1, -1)
		texture:SetPoint("BOTTOMRIGHT", -1, 1)
		texture:SetTexture(0, 0, 0, 1)
		local textureBackground = button:CreateTexture(nil, "BACKGROUND")
		textureBackground:SetPoint("TOPLEFT", 0, 0)
		textureBackground:SetPoint("BOTTOMRIGHT", 0, 0)
		textureBackground:SetTexture(0.4, 0.4, 0.4, 1)

		local textureNormal = button:CreateTexture(nil, "ARTWORK")
		textureNormal:SetPoint("TOPLEFT", 3, -3)
		textureNormal:SetPoint("BOTTOMRIGHT", -3, 3)
		textureNormal:SetTexture(1, 1, 1, 0)
		button:SetNormalTexture(textureNormal)
		local textureHighlight = button:CreateTexture(nil, "ARTWORK")
		textureHighlight:SetPoint("TOPLEFT", 3, -3)
		textureHighlight:SetPoint("BOTTOMRIGHT", -3, 3)
		textureHighlight:SetTexture(1, 1, 1, 0.2)
		button:SetHighlightTexture(textureHighlight)

		if small then
			button:SetNormalFontObject("GameFontNormalSmall")
			button:SetHighlightFontObject("GameFontWhiteSmall")
		else
			button:SetNormalFontObject("GameFontNormal")
			button:SetHighlightFontObject("GameFontHighlight")
		end
	end

	local function TemplateSmallCollapsButton(button, which)
		local name = button:GetName()
		local texture = button:CreateTexture(nil, "BORDER")
		texture:SetPoint("TOPLEFT", 1, -1)
		texture:SetPoint("BOTTOMRIGHT", -1, 1)
		texture:SetTexture(0, 0, 0, 1)
		local textureBackground = button:CreateTexture(name.."BorderColor", "BACKGROUND")
		textureBackground:SetPoint("TOPLEFT", 0, 0)
		textureBackground:SetPoint("BOTTOMRIGHT", 0, 0)
		textureBackground:SetTexture(0.4, 0.4, 0.4, 1)

		local textureNormal = button:CreateTexture(name.."Normal", "ARTWORK")
		textureNormal:SetPoint("TOPLEFT", 4, -4)
		textureNormal:SetPoint("BOTTOMRIGHT", -4, 4)
		if which == 1 then
			textureNormal:SetTexture(IMAGE_Icon) -- GroupScan
			textureNormal:SetTexCoord(0.3125, 0.609375, 0, 0.5625) -- (20/64, 39/64, 0/32, 18/32)
		elseif which == 2 then
			--OLDtextureNormal:SetTexture("Interface\\Buttons\\UI-Panel-ExpandButton-Up")
			--OLD                               0,1              1,1             0,0              1,0
			--OLD                             ULx,ULy          LLx,LLy         URx,URy          LRx,LRy
			--OLD                            6/32,25/32      25/32,25/32      6/32,7/32       25/32,7/32
			--OLDtextureNormal:SetTexCoord(0.1875,0.78125, 0.78125,0.78125, 0.1875,0.21875, 0.78125,0.21875)
			textureNormal:SetTexture(IMAGE_Icon) -- History
			textureNormal:SetTexCoord(0, 0.296875, 0, 0.5625) -- (0/64, 19/64, 0/32, 18/32)
		else
			textureNormal:SetTexture("Interface\\Buttons\\UI-Panel-ExpandButton-Up")
			textureNormal:SetTexCoord(0.1875, 0.78125, 0.21875, 0.78125) -- (6/32, 25/32, 7/32, 25/32)
		end
		button:SetNormalTexture(textureNormal)

		local texturePush = button:CreateTexture(name.."Push", "ARTWORK")
		texturePush:SetPoint("TOPLEFT", 5, -5)
		texturePush:SetPoint("BOTTOMRIGHT", -5, 5)
		if which == 1 then
			texturePush:SetTexture(IMAGE_Icon) -- GroupScan
			texturePush:SetTexCoord(0.3125, 0.609375, 0, 0.5625) -- (20/64, 39/64, 0/32, 18/32)
		elseif which == 2 then
			texturePush:SetTexture(IMAGE_Icon) -- History
			texturePush:SetTexCoord(0, 0.296875, 0, 0.5625) -- (0/64, 19/64, 0/32, 18/32)
		else
			texturePush:SetTexture("Interface\\Buttons\\UI-Panel-ExpandButton-Up")
			texturePush:SetTexCoord(0.1875, 0.78125, 0.21875, 0.78125) -- (6/32, 25/32, 7/32, 25/32)
		end
		button:SetPushedTexture(texturePush)

		local textureDisable = button:CreateTexture(name.."Disable", "ARTWORK")
		textureDisable:SetPoint("TOPLEFT", 4, -4)
		textureDisable:SetPoint("BOTTOMRIGHT", -4, 4)
		if which == 1 then
			textureDisable:SetTexture(IMAGE_Icon) -- GroupScan
			textureDisable:SetTexCoord(0.3125, 0.609375, 0, 0.5625) -- (20/64, 39/64, 0/32, 18/32)
		elseif which == 2 then
			textureDisable:SetTexture(IMAGE_Icon) -- History
			textureDisable:SetTexCoord(0, 0.296875, 0, 0.5625) -- (0/64, 19/64, 0/32, 18/32)
		else
			textureDisable:SetTexture("Interface\\Buttons\\UI-Panel-ExpandButton-Up")
			textureDisable:SetTexCoord(0.1875, 0.78125, 0.21875, 0.78125) -- (6/32, 25/32, 7/32, 25/32)
		end
		button:SetDisabledTexture(textureDisable)

		local shaderSupported = textureDisable:SetDesaturated(true)
		if not shaderSupported then
			textureDisable:SetVertexColor(0.5, 0.5, 0.5)
		end

		local textureHighlight = button:CreateTexture(nil, "ARTWORK")
		textureHighlight:SetPoint("TOPLEFT", 3, -3)
		textureHighlight:SetPoint("BOTTOMRIGHT", -3, 3)
		textureHighlight:SetTexture(1, 1, 1, 0.2)
		button:SetHighlightTexture(textureHighlight)

		button:SetNormalFontObject("GameFontNormal")
		button:SetHighlightFontObject("GameFontHighlight")
	end

	local function TemplateGroupSelectionButton(button)
		local name = button:GetName()
		local texture = button:CreateTexture(name.."Border", "BORDER")
		texture:SetPoint("TOPLEFT", 1, -1)
		texture:SetPoint("BOTTOMRIGHT", -1, 1)
		texture:SetTexture(0, 0, 0, 1)
		local textureBackground = button:CreateTexture(nil, "BACKGROUND")
		textureBackground:SetPoint("TOPLEFT", 0, 0)
		textureBackground:SetPoint("BOTTOMRIGHT", 0, 0)
		textureBackground:SetTexture(0.4, 0.4, 0.4, 1)
	
		local textureNormal = button:CreateTexture(nil, "ARTWORK")
		textureNormal:SetPoint("TOPLEFT", 3, -3)
		textureNormal:SetPoint("BOTTOMRIGHT", -3, 3)
		textureNormal:SetTexture(1, 1, 1, 0)
		button:SetNormalTexture(textureNo1rmal)
		local textureHighlight = button:CreateTexture(nil, "ARTWORK")
		textureHighlight:SetPoint("TOPLEFT", 3, -3)
		textureHighlight:SetPoint("BOTTOMRIGHT", -3, 3)
		textureHighlight:SetTexture(1, 1, 1, 0.2)
		button:SetHighlightTexture(textureHighlight)

		button:SetNormalFontObject("GameFontNormalSmall")
		button:SetHighlightFontObject("GameFontWhiteSmall")
	end

	local function TemplateBorderTRBL(frame) -- TRBL = Top-Right-Bottom-Left
		local texture = frame:CreateTexture(nil, "BORDER")
		texture:SetPoint("TOPLEFT", 1, -1)
		texture:SetPoint("BOTTOMRIGHT", -1, 1)
		texture:SetTexture(0, 0, 0, 1)
		local textureBackground = frame:CreateTexture(nil, "BACKGROUND")
		textureBackground:SetPoint("TOPLEFT", 0, 0)
		textureBackground:SetPoint("BOTTOMRIGHT", 0, 0)
		textureBackground:SetTexture(0.4, 0.4, 0.4, 1)
	end

	local function TemplateBorderRBL(frame) -- RBL = Right-Bottom-Left
		local name = frame:GetName()
		local texture = frame:CreateTexture(name.."Border", "BORDER")
		texture:SetPoint("TOPLEFT", 1, 0)
		texture:SetPoint("BOTTOMLEFT", 0, 1)
		texture:SetPoint("TOPRIGHT", -1, 0)
		texture:SetPoint("BOTTOMRIGHT", -1, -1)
		texture:SetTexture(0, 0, 0, 1)
		local textureBackground = frame:CreateTexture(nil, "BACKGROUND")
		textureBackground:SetPoint("TOPLEFT", 0, 0)
		textureBackground:SetPoint("BOTTOMRIGHT", 0, 0)
		textureBackground:SetTexture(0.4, 0.4, 0.4, 1)
	end

	local function TemplateBorderRL(frame) -- RL = Right-Left
		local texture = frame:CreateTexture(nil, "BORDER")
		texture:SetPoint("TOPLEFT", 1, 0)
		texture:SetPoint("BOTTOMLEFT", 0, 0)
		texture:SetPoint("TOPRIGHT", -1, 0)
		texture:SetPoint("BOTTOMRIGHT", 0, -1)
		texture:SetTexture(0, 0, 0, 0.4)
		local textureBackground = frame:CreateTexture(nil, "BACKGROUND")
		textureBackground:SetPoint("TOPLEFT", 0, 0)
		textureBackground:SetPoint("BOTTOMRIGHT", 0, 0)
		textureBackground:SetTexture(0.4, 0.4, 0.4, 1)
	end

	local function TemplateStats(button, collapse, level)
		local collapseButtonWidth = 0
		local nameButtonWidth = 200
		local nameButtonSpace = 0
		local achievementButtonWidth = 80
		local achievementButtonSpace = 200
		local statisticButtonWidth = 60
		local statisticButtonSpace = 280
		local bossestotalButtonWidth = 50
		local bossestotalButtonSpace = 340
		local font = "GameFontWhiteSmall"

		if level == "title" then
			font = "GameFontNormalSmall"
			if collapse then
				nameButtonWidth = 190
				nameButtonSpace = 10
			end
		elseif level == "level1" then
			if collapse then
				nameButtonWidth = 190
				nameButtonSpace = 10
			end
		elseif level == "level2" then
			nameButtonWidth = 340
			nameButtonSpace = 10
			if collapse then
				nameButtonWidth = 190
				nameButtonSpace = 20
				collapseButtonWidth = 10
			end
		elseif level == "level3" then
			nameButtonWidth = 320
			nameButtonSpace = 20
			if collapse then
				--nameButtonWidth = 220
				nameButtonSpace = 30
				collapseButtonWidth = 20
			end
		elseif level == "level4" then
			nameButtonWidth = 310
			nameButtonSpace = 30
			if collapse then
				--nameButtonWidth = 220
				nameButtonSpace = 30
				collapseButtonWidth = 20
			end
		end

		button:SetWidth(390)
		button:SetHeight(12)
		local name = button:GetName()
		if collapse then
			local collapseButton = button:CreateTexture(name.."Collapse", "ARTWORK")
			collapseButton:SetWidth(8)
			collapseButton:SetHeight(8)
			collapseButton:SetPoint("TOPLEFT", collapseButtonWidth, -2)
			collapseButton:SetTexture("Interface\\Buttons\\UI-PlusButton-Up")
			collapseButton:SetTexCoord(0.25, 0.6875, 0.3125, 0.75) -- (4/16, 11/16, 5/16, 12/16)
		end
		local nameButton = button:CreateFontString(name.."Name", "OVERLAY", font)
		nameButton:SetWidth(nameButtonWidth)
		nameButton:SetHeight(12)
		nameButton:SetPoint("LEFT", nameButtonSpace, 0)
		nameButton:SetJustifyH("LEFT")
		local achievementButton = button:CreateFontString(name.."Achievement", "OVERLAY", font)
		achievementButton:SetWidth(achievementButtonWidth)
		achievementButton:SetHeight(12)
		achievementButton:SetPoint("LEFT", achievementButtonSpace, 0)
		achievementButton:SetJustifyH("RIGHT")
		local statisticButton = button:CreateFontString(name.."Statistic", "OVERLAY", font)
		statisticButton:SetWidth(statisticButtonWidth)
		statisticButton:SetHeight(12)
		statisticButton:SetPoint("LEFT", statisticButtonSpace, 0)
		statisticButton:SetJustifyH("RIGHT")
		local bossestotalButton = button:CreateFontString(name.."BossesTotal", "OVERLAY", font)
		bossestotalButton:SetWidth(bossestotalButtonWidth)
		bossestotalButton:SetHeight(12)
		bossestotalButton:SetPoint("LEFT", bossestotalButtonSpace, 0)
		bossestotalButton:SetJustifyH("RIGHT")
		if level ~= "title" and level ~= "summary" then
			local textureHighlight = button:CreateTexture(name.."Highlight", "ARTWORK")
			textureHighlight:SetPoint("TOPLEFT", 0, 0)
			textureHighlight:SetPoint("BOTTOMRIGHT", 0, 0)
			textureHighlight:SetTexture(1, 1, 1, 0.2)
			button:SetHighlightTexture(textureHighlight)
			local texture = button:CreateTexture(name.."Texture", "BACKGROUND")
			texture:SetPoint("TOPLEFT", nameButtonSpace, -5)
			texture:SetPoint("BOTTOMRIGHT", 0, 5)
			texture:SetTexture(0.06, 0.06, 0.06, 1)
			button:SetNormalTexture(texture)
		end
	end

	local function TemplateHistoryList(button, buttonid)
		local name = button:GetName()
		button:SetWidth(220)
		button:SetHeight(12)

		local deleteButton = CreateFrame("Button", nil, button)
		deleteButton:SetWidth(9)
		deleteButton:SetHeight(11)
		deleteButton:SetPoint("LEFT", button, "LEFT", 5, 0)
		local deleteButtonTexture = deleteButton:CreateTexture(nil, "ARTWORK")
		deleteButtonTexture:SetPoint("LEFT", 0, 0)
		deleteButtonTexture:SetWidth(9)
		deleteButtonTexture:SetHeight(11)
		deleteButtonTexture:SetTexture(IMAGE_Icon) -- trashcan
		deleteButtonTexture:SetTexCoord(0, 0.140625, 0.59375, 0.9375) -- (0/64,  9/64, 19/32, 30/32)
		deleteButtonTexture:SetAlpha(0.5)
		deleteButton:SetScript("OnEnter", function() deleteButtonTexture:SetAlpha(1) end)
		deleteButton:SetScript("OnLeave", function() deleteButtonTexture:SetAlpha(0.5) end)
		deleteButton:SetScript("OnClick", function() CharScanner:DeleteCharacter(button.guid, buttonid) end)

		local texture = button:CreateTexture(name.."Arrow", "BACKGROUND")
		texture:SetWidth(6)
		texture:SetHeight(10)
		texture:SetTexture("Interface\\Buttons\\UI-SpellbookIcon-NextPage-Up")
		texture:SetTexCoord(0.34375, 0.625, 0.28125, 0.75) -- (11/32, 20/32, 9/32, 24/32)
		texture:SetPoint("LEFT", deleteButton, "RIGHT", 2, 0)
		texture:Hide()

		local iconButton = button:CreateTexture(name.."Icon", "ARTWORK")
		iconButton:SetWidth(12)
		iconButton:SetHeight(12)
		iconButton:SetPoint("LEFT", button, "LEFT", 22, 0)
		iconButton:SetTexture(classimg)

		local nameButton = button:CreateFontString(name.."Name", "ARTWORK", "GameFontNormalSmall")
		nameButton:SetWidth(120)
		nameButton:SetHeight(12)
		nameButton:SetPoint("LEFT", iconButton, "RIGHT", 2, 0)
		nameButton:SetJustifyH("LEFT")

		local nameTime = button:CreateFontString(name.."Time", "ARTWORK", "GameFontWhiteSmall")
		nameTime:SetWidth(40)
		nameTime:SetHeight(12)
		nameTime:SetPoint("LEFT", nameButton, "RIGHT", 2, 0)
		nameTime:SetJustifyH("LEFT")

		local textureHighlight = button:CreateTexture(name.."Highlight", "ARTWORK")
		textureHighlight:SetPoint("TOPLEFT", 5, 0)
		textureHighlight:SetPoint("BOTTOMRIGHT", -5, 0)
		textureHighlight:SetTexture(1, 1, 1, 0.2)
		button:SetHighlightTexture(textureHighlight)

		button:SetScript("OnClick", function() CharScanner:PreShowSavedData(button.guid) end)
	end

	local function TemplateTalentButton(button)
		button:SetWidth(25)
		button:SetHeight(25)
		local name = button:GetName()

		local texture = button:CreateTexture(name.."IconTexture", "BORDER")
		texture:SetWidth(25)
		texture:SetHeight(25)
		texture:SetPoint("TOPLEFT", 0, 0)

		local texture = button:CreateTexture(name.."RankBorder", "OVERLAY")
		texture:SetWidth(14)
		texture:SetHeight(14)
		texture:SetPoint("BOTTOMRIGHT", name.."IconTexture", "BOTTOMRIGHT", 1, -1)
		texture:SetTexture("Interface\\TalentFrame\\TalentFrame-RankBorder")
		texture:SetTexCoord(0.28125, 0.71875, 0.28125, 0.6875) -- (9/32, 23/32, 9/32, 22/32)
		texture:SetAlpha(0.8)

		local text = button:CreateFontString(name.."Rank", "OVERLAY", "GameFontNormalSmall")
		text:SetPoint("CENTER", name.."RankBorder", "CENTER", 0, 0)
		text:SetJustifyH("RIGHT")

		local textureHighlight = button:CreateTexture(nil, "ARTWORK")
		textureHighlight:SetWidth(25)
		textureHighlight:SetHeight(25)
		textureHighlight:SetPoint("TOPLEFT", 0, 0)
		textureHighlight:SetTexture("Interface\\Buttons\\ButtonHilight-Square")
		button:SetHighlightTexture(textureHighlight)
	end

	local function TemplateGearButton(button, pos)
		button:SetWidth(25)
		button:SetHeight(25)
		local name = button:GetName()

		local texture = button:CreateTexture(name.."IconTexture", "BORDER")
		texture:SetWidth(25)
		texture:SetHeight(25)
		texture:SetPoint("TOPLEFT", 0, 0)

		local text = button:CreateFontString(name.."Count", "OVERLAY", "GameFontNormalSmall")
		if pos == 1 then -- LEFT
			text:SetPoint("RIGHT", texture, "LEFT", -5, 0)
		elseif pos == 2 then -- RIGHT
			text:SetPoint("LEFT", texture, "RIGHT", 5, 0)
		elseif pos == 3 then -- BOTTOM
			text:SetPoint("TOP", texture, "BOTTOM", 0, -5)
		end

		local textureHighlight = button:CreateTexture(name.."Highlight", "ARTWORK")
		textureHighlight:SetWidth(25)
		textureHighlight:SetHeight(25)
		textureHighlight:SetPoint("TOPLEFT", 0, 0)
		textureHighlight:SetTexture("Interface\\Buttons\\ButtonHilight-Square")
		button:SetHighlightTexture(textureHighlight)
	end
	-- frame/button templates END

	local mainframe = CreateFrame("Frame", "CharScanner_MainFrame", UIParent)
	mainframe:EnableMouse(true)
	mainframe:SetMovable(true)
	mainframe:SetToplevel(true)
	mainframe:SetFrameStrata("DIALOG")
	mainframe:SetClampedToScreen(true)
	mainframe:Hide()
	mainframe:SetWidth(400)
	mainframe:SetHeight(30)
	mainframe:SetScript("OnShow", function() CharScanner:MainFrameShow() end)
	mainframe:SetScript("OnHide", function() CharScanner:MainFrameHide() end)

	-- Headline Frame START
	local headline = CreateFrame("Frame", "CharScanner_MainFrame_Headline", CharScanner_MainFrame)
	TemplateBorderTRBL(headline)
	headline:SetWidth(400)
	headline:SetHeight(30)
	headline:SetPoint("TOPLEFT", CharScanner_MainFrame, "TOPLEFT", 0, 0)
	headline:EnableMouse(true)

	local headlineButtonScan = CreateFrame("Button", "CharScanner_MainFrame_Headline_ButtonScan", headline)
	TemplateHeadlineButton(headlineButtonScan)
	headlineButtonScan:SetText(L["SCAN Target"])
	headlineButtonScan:SetWidth( ceil(CharScanner_MainFrame_Headline_ButtonScan:GetTextWidth()) + 20 )
	headlineButtonScan:SetHeight(22)
	headlineButtonScan:SetPoint("LEFT", headline, "LEFT", 4, 0)
	headlineButtonScan:SetScript("OnClick", function() CharScanner:ScanUnit("target") end)

	local headlineButtonGroupScan = CreateFrame("Button", "CharScanner_MainFrame_Headline_ButtonGroupScan", headline)
	TemplateSmallCollapsButton(headlineButtonGroupScan, 1)
	headlineButtonGroupScan:SetWidth(22)
	headlineButtonGroupScan:SetHeight(22)
	headlineButtonGroupScan:SetPoint("LEFT", headlineButtonScan, "RIGHT", 4, 0)
	headlineButtonGroupScan:Disable()
	CharScanner_MainFrame_Headline_ButtonGroupScanBorderColor:SetTexture(0.2, 0.2, 0.2, 1)
	headlineButtonGroupScan:SetScript("OnClick", function() CharScanner:ScanGroup() end)

	local headlineButtonCollapse = CreateFrame("Button", "CharScanner_MainFrame_Headline_ButtonCollapse", headline)
	TemplateSmallCollapsButton(headlineButtonCollapse, 0)
	headlineButtonCollapse:SetWidth(22)
	headlineButtonCollapse:SetHeight(22)
	headlineButtonCollapse:SetPoint("LEFT", headlineButtonGroupScan, "RIGHT", 4, 0)
	headlineButtonCollapse:Disable()
	CharScanner_MainFrame_Headline_ButtonCollapseBorderColor:SetTexture(0.2, 0.2, 0.2, 1)
	headlineButtonCollapse:SetScript("OnClick", function() CharScanner:ScanFrameToggle() end)

	local headlineButtonHistory = CreateFrame("Button", "CharScanner_MainFrame_Headline_ButtonHistory", headline)
	TemplateSmallCollapsButton(headlineButtonHistory, 2)
	headlineButtonHistory:SetWidth(22)
	headlineButtonHistory:SetHeight(22)
	headlineButtonHistory:SetPoint("LEFT", headlineButtonCollapse, "RIGHT", 4, 0)
	headlineButtonHistory:Disable()
	CharScanner_MainFrame_Headline_ButtonHistoryBorderColor:SetTexture(0.2, 0.2, 0.2, 1)
	headlineButtonHistory:SetScript("OnClick", function() CharScanner:HistoryFrameToggle() end)

	local headlineButtonX = CreateFrame("Button", "CharScanner_MainFrame_Headline_ButtonX", headline)
	TemplateHeadlineButton(headlineButtonX)
	headlineButtonX:SetWidth(22)
	headlineButtonX:SetHeight(22)
	headlineButtonX:SetPoint("RIGHT", headline, "RIGHT", -4, 0)
	local headlineButtonXTextureNormal = headlineButtonX:CreateTexture(nil, "ARTWORK")
	headlineButtonXTextureNormal:SetPoint("TOPLEFT", 4, -4)
	headlineButtonXTextureNormal:SetPoint("BOTTOMRIGHT", -4, 4)
	headlineButtonXTextureNormal:SetTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Up")
	headlineButtonXTextureNormal:SetTexCoord(0.25, 0.6875, 0.3125, 0.6875) -- (8/32, 22/32, 10/32, 22/32)
	headlineButtonX:SetNormalTexture(headlineButtonXTextureNormal)
	local headlineButtonXTexturePush = headlineButtonX:CreateTexture(nil, "ARTWORK")
	headlineButtonXTexturePush:SetPoint("TOPLEFT", 5, -5)
	headlineButtonXTexturePush:SetPoint("BOTTOMRIGHT", -5, 5)
	headlineButtonXTexturePush:SetTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Up")
	headlineButtonXTexturePush:SetTexCoord(0.25, 0.6875, 0.3125, 0.6875) -- (8/32, 22/32, 10/32, 22/32)
	headlineButtonX:SetPushedTexture(headlineButtonXTexturePush)
	headlineButtonX:SetScript("OnClick", function() CharScanner:MainFrameToggle() end)

	local headlineButtonOptions = CreateFrame("Button", "CharScanner_MainFrame_Headline_ButtonOptions", headline)
	TemplateHeadlineButton(headlineButtonOptions)
	headlineButtonOptions:SetWidth(22)
	headlineButtonOptions:SetHeight(22)
	headlineButtonOptions:SetPoint("RIGHT", headlineButtonX, "LEFT", -4, 0)
	local headlineButtonOptionsTextureNormal = headlineButtonOptions:CreateTexture(nil, "ARTWORK")
	headlineButtonOptionsTextureNormal:SetPoint("TOPLEFT", 4, -4)
	headlineButtonOptionsTextureNormal:SetPoint("BOTTOMRIGHT", -4, 4)
	headlineButtonOptionsTextureNormal:SetTexture("Interface\\GossipFrame\\BinderGossipIcon")
	headlineButtonOptions:SetNormalTexture(headlineButtonOptionsTextureNormal)
	local headlineButtonOptionsTexturePush = headlineButtonOptions:CreateTexture(nil, "ARTWORK")
	headlineButtonOptionsTexturePush:SetPoint("TOPLEFT", 5, -5)
	headlineButtonOptionsTexturePush:SetPoint("BOTTOMRIGHT", -5, 5)
	headlineButtonOptionsTexturePush:SetTexture("Interface\\GossipFrame\\BinderGossipIcon")
	headlineButtonOptions:SetPushedTexture(headlineButtonOptionsTexturePush)
	headlineButtonOptions:SetScript("OnClick", function() CharScanner:MainFrameToggle() InterfaceOptionsFrame_OpenToCategory("CharScanner") end)

	local headlineButtonMover = CreateFrame("Button", "CharScanner_MainFrame_Headline_ButtonMover", headline)
	headlineButtonMover:SetWidth(22)
	headlineButtonMover:SetHeight(22)
	headlineButtonMover:SetPoint("TOPLEFT", headlineButtonHistory, "TOPRIGHT", 4, 0)
	headlineButtonMover:SetPoint("BOTTOMRIGHT", headlineButtonOptions, "BOTTOMLEFT", -4, 0)
	local textureMover = headlineButtonMover:CreateTexture(nil, "BACKGROUND")
	textureMover:SetWidth(1)
	textureMover:SetHeight(1)
	textureMover:SetPoint("TOPLEFT", headlineButtonMover, "TOPLEFT", 0, 0)
	textureMover:SetPoint("BOTTOMRIGHT", headlineButtonMover, "BOTTOMRIGHT", 0, 0)
	textureMover:SetTexture(1, 1, 1, 0)
	headlineButtonMover:SetScript("OnEnter", function() textureMover:SetTexture(1, 1, 1, 0.075) end)
	headlineButtonMover:SetScript("OnLeave", function() textureMover:SetTexture(1, 1, 1, 0) end)
	headlineButtonMover:SetScript("OnMouseDown", function() CharScanner_MainFrame:StartMoving() end)
	headlineButtonMover:SetScript("OnMouseUp", function() CharScanner_MainFrame:StopMovingOrSizing() CharScanner:SaveMainFramePos() end)
	-- Headline Frame END

	-- Base Character Information Frame START
	local base = CreateFrame("Frame", "CharScanner_MainFrame_Base", mainframe)
	TemplateBorderRBL(base)
	base:SetWidth(400)
	base:SetHeight(38)
	base:SetPoint("TOPLEFT", headline, "BOTTOMLEFT", 0, 0)
	base:EnableMouse(true)
	base:SetScript("OnShow", function() CharScanner:CheckStatusFramePosition() end)
	base:Hide()

	local baseClassIcon = base:CreateTexture("CharScanner_MainFrame_Base_ClassIcon", "ARTWORK")
	baseClassIcon:SetWidth(28)
	baseClassIcon:SetHeight(28)
	baseClassIcon:SetPoint("LEFT", base, "LEFT", 5, 1)

	local baseFactionIcon = base:CreateTexture("CharScanner_MainFrame_Base_FactionIcon", "ARTWORK")
	baseFactionIcon:SetWidth(18)
	baseFactionIcon:SetHeight(18)
	baseFactionIcon:SetPoint("LEFT", baseClassIcon, "RIGHT", 0, 0)

	local baseName = base:CreateFontString("CharScanner_MainFrame_Base_Name", "ARTWORK", "GameFontNormal")
	baseName:SetPoint("TOPLEFT", base, "TOPLEFT", 50, -2)
	baseName:SetJustifyH("LEFT")
	baseName:SetTextColor(1, 1, 1, 1) -- TEST

	local baseLevel = base:CreateFontString("CharScanner_MainFrame_Base_Level", "ARTWORK", "GameFontNormalSmall")
	baseLevel:SetPoint("TOPLEFT", baseName, "BOTTOMLEFT", 0, 0)
	baseLevel:SetJustifyH("LEFT")
	baseLevel:SetTextColor(1, 1, 1, 1) -- TEST

	local baseGuild = base:CreateFontString("CharScanner_MainFrame_Base_Guild", "ARTWORK", "GameFontNormalSmall")
	baseGuild:SetPoint("TOPLEFT", baseLevel, "BOTTOMLEFT", 0, 0)
	baseGuild:SetJustifyH("LEFT")
	baseGuild:SetTextColor(1, 1, 1, 1) -- TEST
	-- Base Character Information Frame END

	-- Talent Frame START
	local talent = CreateFrame("Frame", "CharScanner_MainFrame_Talent", mainframe)
	TemplateBorderRBL(talent)
	talent:SetWidth(400)
	talent:SetHeight(28)
	talent:SetPoint("TOPLEFT", base, "BOTTOMLEFT", 0, 0)
	talent:EnableMouse(true)
	talent:Hide()

	local talentTitle = talent:CreateFontString("CharScanner_MainFrame_Talent_Title", "ARTWORK", "GameFontNormalSmall")
	talentTitle:SetPoint("TOPLEFT", talent, "TOPLEFT", 5, -2)
	talentTitle:SetJustifyH("LEFT")

	local talentGroup1Icon = talent:CreateTexture("CharScanner_MainFrame_Talent_TalentGroup1Icon", "ARTWORK")
	talentGroup1Icon:SetWidth(12)
	talentGroup1Icon:SetHeight(12)
	talentGroup1Icon:SetPoint("LEFT", talentTitle, "RIGHT", 5, 0)
	talentGroup1Icon:SetTexCoord(0.078125, 0.921875, 0.078125, 0.921875) -- (5/64, 59/64, 5/64, 59/64)
	local talentGroup1Text = talent:CreateFontString("CharScanner_MainFrame_Talent_TalentGroup1Text", "ARTWORK", "GameFontNormalSmall")
	talentGroup1Text:SetPoint("LEFT", talentGroup1Icon, "RIGHT", 5, 0)
	talentGroup1Text:SetJustifyH("LEFT")

	local talentGroup2Icon = talent:CreateTexture("CharScanner_MainFrame_Talent_TalentGroup2Icon", "ARTWORK")
	talentGroup2Icon:SetWidth(12)
	talentGroup2Icon:SetHeight(12)
	talentGroup2Icon:SetPoint("TOPLEFT", talentGroup1Icon, "BOTTOMLEFT", 0, 0)
	talentGroup2Icon:SetTexCoord(0.078125, 0.921875, 0.078125, 0.921875) -- (5/64, 59/64, 5/64, 59/64)
	local talentGroup2Text = talent:CreateFontString("CharScanner_MainFrame_Talent_TalentGroup2Text", "ARTWORK", "GameFontNormalSmall")
	talentGroup2Text:SetPoint("LEFT", talentGroup2Icon, "RIGHT", 5, 0)
	talentGroup2Text:SetJustifyH("LEFT")

	local talentButtonCollapse = CreateFrame("Button", "CharScanner_MainFrame_Talent_ButtonCollapse", talent)
	TemplateSmallCollapsButton(talentButtonCollapse, 0)
	talentButtonCollapse:SetWidth(22)
	talentButtonCollapse:SetHeight(22)
	talentButtonCollapse:SetPoint("RIGHT", talent, "RIGHT", 0, 0)
	talentButtonCollapse:SetScript("OnClick", function()
		CharScanner_MainFrame_TalentTree_GroupSelection_Button2:EnableMouse(true) -- fast MouseOver Button2 hackfix
		CharScanner:TalentTreeFrameToggle()
	end)
	talentButtonCollapse:SetScript("OnEnter", function()
		if not TalentTreeFrameIsShown then
			CharScanner_MainFrame_TalentTree:Show()
			CharScanner_MainFrame_TalentTree_GroupSelection_Button2:EnableMouse(false) -- fast MouseOver Button2 hackfix
			CharScanner_MainFrame_TalentBorder:SetPoint("TOPLEFT", 1, 0)
			CharScanner_MainFrame_TalentBorder:SetPoint("BOTTOMLEFT", 0, 0)
			CharScanner_MainFrame_TalentBorder:SetPoint("TOPRIGHT", -1, 0)
			CharScanner_MainFrame_TalentBorder:SetPoint("BOTTOMRIGHT", 0, 1)
		end
	end)
	talentButtonCollapse:SetScript("OnLeave", function()
		if not TalentTreeFrameIsShown then
			CharScanner_MainFrame_TalentTree:Hide()
			CharScanner_MainFrame_TalentTree_GroupSelection_Button2:EnableMouse(true) -- fast MouseOver Button2 hackfix
			CharScanner_MainFrame_TalentBorder:SetPoint("TOPLEFT", 1, 0)
			CharScanner_MainFrame_TalentBorder:SetPoint("BOTTOMLEFT", 0, 1)
			CharScanner_MainFrame_TalentBorder:SetPoint("TOPRIGHT", -1, 0)
			CharScanner_MainFrame_TalentBorder:SetPoint("BOTTOMRIGHT", -1, -1)
		end
	end)

	-- TalentTree Frame START
	local TalentTree = CreateFrame("Frame", "CharScanner_MainFrame_TalentTree", talent)
	TemplateBorderRBL(TalentTree)
	TalentTree:SetToplevel(true)
	TalentTree:SetWidth(400)
	TalentTree:SetHeight(371) -- (11*25) + (12*4) + 16 + (2*16)
	TalentTree:SetPoint("TOPLEFT", talent, "BOTTOMLEFT", 0, 0)
	TalentTree:EnableMouse(true)
	TalentTree:Hide()

	local GroupSelection = CreateFrame("Frame", "CharScanner_MainFrame_TalentTree_GroupSelection", TalentTree)
	TemplateBorderRBL(GroupSelection)
	GroupSelection:SetWidth(400)
	GroupSelection:SetHeight(16)
	GroupSelection:SetPoint("TOPLEFT", TalentTree, "TOPLEFT", 0, 0)

	local GroupSelection_Button1 = CreateFrame("Button", "CharScanner_MainFrame_TalentTree_GroupSelection_Button1", GroupSelection)
	TemplateGroupSelectionButton(GroupSelection_Button1)
	GroupSelection_Button1:SetWidth(194)
	GroupSelection_Button1:SetHeight(16)
	GroupSelection_Button1:SetPoint("BOTTOMLEFT", GroupSelection, "BOTTOMLEFT", 4, 0)
	GroupSelection_Button1:SetText(TALENT_SPEC_PRIMARY)
	GroupSelection_Button1:Hide()

	local GroupSelection_Button2 = CreateFrame("Button", "CharScanner_MainFrame_TalentTree_GroupSelection_Button2", GroupSelection)
	TemplateGroupSelectionButton(GroupSelection_Button2)
	GroupSelection_Button2:SetWidth(194)
	GroupSelection_Button2:SetHeight(16)
	GroupSelection_Button2:SetPoint("BOTTOMLEFT", GroupSelection, "BOTTOMLEFT", 202, 0)
	GroupSelection_Button2:SetText(TALENT_SPEC_SECONDARY)
	GroupSelection_Button2:Hide()

	local TalentTree1 = CreateFrame("Frame", "CharScanner_MainFrame_TalentTree1", TalentTree)
	TalentTree1:SetWidth(120)
	TalentTree1:SetHeight(323) -- (11*25) + (12*4)
	TalentTree1:SetPoint("TOPLEFT", GroupSelection, "BOTTOMLEFT", 10, -4)
	local TalentTree1Name = TalentTree1:CreateFontString("CharScanner_MainFrame_TalentTree1Name", "ARTWORK", "GameFontWhiteSmall")
	TalentTree1Name:SetPoint("TOP", TalentTree1, "TOP", 0, 0)
	TalentTree1Name:SetJustifyH("CENTER")
	local TalentTree1Points = TalentTree1:CreateFontString("CharScanner_MainFrame_TalentTree1Points", "ARTWORK", "GameFontWhiteSmall")
	TalentTree1Points:SetPoint("TOP", TalentTree1Name, "BOTTOM", 0, 0)
	TalentTree1Points:SetJustifyH("CENTER")

	local TalentTree2 = CreateFrame("Frame", "CharScanner_MainFrame_TalentTree2", TalentTree)
	TalentTree2:SetWidth(120)
	TalentTree2:SetHeight(323) -- (11*25) + (12*4)
	TalentTree2:SetPoint("TOPLEFT", GroupSelection, "BOTTOMLEFT", 140, -4)
	local TalentTree2Name = TalentTree2:CreateFontString("CharScanner_MainFrame_TalentTree2Name", "ARTWORK", "GameFontWhiteSmall")
	TalentTree2Name:SetPoint("TOP", TalentTree2, "TOP", 0, 0)
	TalentTree2Name:SetJustifyH("CENTER")
	local TalentTree2Points = TalentTree2:CreateFontString("CharScanner_MainFrame_TalentTree2Points", "ARTWORK", "GameFontWhiteSmall")
	TalentTree2Points:SetPoint("TOP", TalentTree2Name, "BOTTOM", 0, 0)
	TalentTree2Points:SetJustifyH("CENTER")

	local TalentTree3 = CreateFrame("Frame", "CharScanner_MainFrame_TalentTree3", TalentTree)
	TalentTree3:SetWidth(120)
	TalentTree3:SetHeight(323) -- (11*25) + (12*4)
	TalentTree3:SetPoint("TOPLEFT", GroupSelection, "BOTTOMLEFT", 270, -4)
	local TalentTree3Name = TalentTree3:CreateFontString("CharScanner_MainFrame_TalentTree3Name", "ARTWORK", "GameFontWhiteSmall")
	TalentTree3Name:SetPoint("TOP", TalentTree3, "TOP", 0, 0)
	TalentTree3Name:SetJustifyH("CENTER")
	local TalentTree3Points = TalentTree3:CreateFontString("CharScanner_MainFrame_TalentTree3Points", "ARTWORK", "GameFontWhiteSmall")
	TalentTree3Points:SetPoint("TOP", TalentTree3Name, "BOTTOM", 0, 0)
	TalentTree3Points:SetJustifyH("CENTER")

	local numColumn = 4
	local tier = 1
	local column = 1
	for tab = 1, myMAX_TALENT_TABS do
		for talent = 1, myMAX_NUM_TALENTS do
			local button
			if talent == 1 then
				tier = 1
				column = 1
				button = CreateFrame("Button", "CharScanner_MainFrame_TalentTree_Slot"..tab.."_"..tier.."_"..column, TalentTree)
				button:SetPoint("TOPLEFT", "CharScanner_MainFrame_TalentTree"..tab, "TOPLEFT", 4, -28) -- -(4 + 2*12) font height
			elseif mod(talent, numColumn) == 1 then
				tier = tier + 1
				column = 1
				button = CreateFrame("Button", "CharScanner_MainFrame_TalentTree_Slot"..tab.."_"..tier.."_"..column, TalentTree)
				button:SetPoint("TOP", "CharScanner_MainFrame_TalentTree_Slot"..tab.."_"..(tier-1).."_"..(column), "BOTTOM", 0, -4)
			else
				column = column + 1
				button = CreateFrame("Button", "CharScanner_MainFrame_TalentTree_Slot"..tab.."_"..tier.."_"..column, TalentTree)
				button:SetPoint("LEFT", "CharScanner_MainFrame_TalentTree_Slot"..tab.."_"..tier.."_"..(column-1), "RIGHT", 4, 0)
			end
			TemplateTalentButton(button)
		end
	end
	-- TalentTree Frame END
	-- Talent Frame END

	-- Equipment Frame START
	local equipment = CreateFrame("Frame", "CharScanner_MainFrame_Equipment", mainframe)
	TemplateBorderRBL(equipment)
	equipment:SetWidth(400)
	equipment:SetHeight(28)
	equipment:SetPoint("TOPLEFT", talent, "BOTTOMLEFT", 0, 0)
	equipment:EnableMouse(true)
	equipment:Hide()

	local equipmentTitle = equipment:CreateFontString("CharScanner_MainFrame_Equipment_Title", "ARTWORK", "GameFontNormalSmall")
	equipmentTitle:SetPoint("TOPLEFT", equipment, "TOPLEFT", 5, -2)
	equipmentTitle:SetJustifyH("LEFT")
	local equipmentText1 = equipment:CreateFontString("CharScanner_MainFrame_Equipment_Text1", "ARTWORK", "GameFontWhiteSmall")
	equipmentText1:SetPoint("LEFT", equipmentTitle, "RIGHT", 5, 0)
	equipmentText1:SetJustifyH("LEFT")
	local equipmentText2 = equipment:CreateFontString("CharScanner_MainFrame_Equipment_Text2", "ARTWORK", "GameFontWhiteSmall")
	equipmentText2:SetPoint("TOPLEFT", equipmentTitle, "TOPRIGHT", 5, -12)
	equipmentText2:SetJustifyH("LEFT")
	
	local equipmentButtonCollapse = CreateFrame("Button", "CharScanner_MainFrame_Equipment_ButtonCollapse", equipment)
	TemplateSmallCollapsButton(equipmentButtonCollapse, 0)
	equipmentButtonCollapse:SetWidth(22)
	equipmentButtonCollapse:SetHeight(22)
	equipmentButtonCollapse:SetPoint("RIGHT", equipment, "RIGHT", 0, 0)
	equipmentButtonCollapse:SetScript("OnClick", function()
		CharScanner:GearFrameToggle()
	end)
	equipmentButtonCollapse:SetScript("OnEnter", function()
		if not GearFrameIsShown then
			CharScanner_MainFrame_Gear:Show()
			CharScanner_MainFrame_EquipmentBorder:SetPoint("TOPLEFT", 1, 0)
			CharScanner_MainFrame_EquipmentBorder:SetPoint("BOTTOMLEFT", 0, 0)
			CharScanner_MainFrame_EquipmentBorder:SetPoint("TOPRIGHT", -1, 0)
			CharScanner_MainFrame_EquipmentBorder:SetPoint("BOTTOMRIGHT", 0, 1)
		end
	end)
	equipmentButtonCollapse:SetScript("OnLeave", function()
		if not GearFrameIsShown then
			CharScanner_MainFrame_Gear:Hide()
			CharScanner_MainFrame_EquipmentBorder:SetPoint("TOPLEFT", 1, 0)
			CharScanner_MainFrame_EquipmentBorder:SetPoint("BOTTOMLEFT", 0, 1)
			CharScanner_MainFrame_EquipmentBorder:SetPoint("TOPRIGHT", -1, 0)
			CharScanner_MainFrame_EquipmentBorder:SetPoint("BOTTOMRIGHT", -1, -1)
		end
	end)
	-- Equipment Frame END

	-- Gear Frame START
	local gear = CreateFrame("Frame", "CharScanner_MainFrame_Gear", equipment)
	TemplateBorderRBL(gear)
	gear:SetToplevel(true)
	gear:SetWidth(400)
	gear:SetHeight(250)
	gear:SetPoint("TOPLEFT", equipment, "BOTTOMLEFT", 0, 0)
	gear:EnableMouse(true)
	gear:Hide()

	for i = 1, #slotButton do
		local button = CreateFrame("Button", "CharScanner_MainFrame_Gear_Slot"..i, gear)
		if i == 1 then
			TemplateGearButton(button, 1)
			button:SetPoint("TOPLEFT", gear, "TOPLEFT", 112.5, -10)
		else
			if i >= 2 and i <= 8 then
				TemplateGearButton(button, 1)
				button:SetPoint("TOPLEFT", "CharScanner_MainFrame_Gear_Slot"..(i-1), "BOTTOMLEFT", 0, 0)
			else
				if i == 9 then
					TemplateGearButton(button, 3)
					button:SetPoint("TOPLEFT", "CharScanner_MainFrame_Gear_Slot"..(i-1), "TOPRIGHT", 25, -10)
				else
					if i >= 10 and i <= 12 then
						TemplateGearButton(button, 3)
						button:SetPoint("TOPLEFT", "CharScanner_MainFrame_Gear_Slot"..(i-1), "TOPRIGHT", 0, 0)
						if i == 12 then
							button:SetScale(0.75)
						end
					else
						if i == 13 then
							TemplateGearButton(button, 2)
							button:SetPoint("TOPLEFT", "CharScanner_MainFrame_Gear_Slot"..(i-2), "TOPRIGHT", 25, 10)
						else
							TemplateGearButton(button, 2)
							button:SetPoint("BOTTOMLEFT", "CharScanner_MainFrame_Gear_Slot"..(i-1), "TOPLEFT", 0, 0)
						end
					end
				end
			end
		end
		_G["CharScanner_MainFrame_Gear_Slot"..i.."IconTexture"]:SetTexture(slotButton[i].texture)
		_G["CharScanner_MainFrame_Gear_Slot"..i.."Count"]:Hide()
	end
	-- disabled slots
	CharScanner_MainFrame_Gear_Slot6Highlight:Hide()
	CharScanner_MainFrame_Gear_Slot6IconTexture:SetAlpha(0.4)
	CharScanner_MainFrame_Gear_Slot7Highlight:Hide()
	CharScanner_MainFrame_Gear_Slot7IconTexture:SetAlpha(0.4)
	CharScanner_MainFrame_Gear_Slot12Highlight:Hide()
	CharScanner_MainFrame_Gear_Slot12IconTexture:SetAlpha(0.4)

	local statsButton = CreateFrame("Button", "CharScanner_MainFrame_Equipment_StatsButton", gear)
	TemplateHeadlineButton(statsButton)
	statsButton:SetText(L["Base Stats"])
	statsButton:SetWidth( ceil(CharScanner_MainFrame_Equipment_StatsButton:GetTextWidth()) + 20 )
	statsButton:SetHeight(22)
	statsButton:SetPoint("TOP", gear, "TOP", 0, -20)
	statsButton:SetScript("OnEnter", function() CharScanner_MainFrame_Equipment_StatsFrame:Show() end)
	statsButton:SetScript("OnLeave", function() CharScanner_MainFrame_Equipment_StatsFrame:Hide() end)
	local statsframe = CreateFrame("Frame", "CharScanner_MainFrame_Equipment_StatsFrame", gear)
	TemplateBorderTRBL(statsframe)
	statsframe:SetToplevel(true)
	statsframe:SetPoint("TOP", statsButton, "BOTTOM", 0, 0)
	statsframe:EnableMouse(true)
	statsframe:Hide()
	local statsText = statsframe:CreateFontString("CharScanner_MainFrame_Equipment_StatsText", "ARTWORK", "GameFontWhiteSmall")
	statsText:SetPoint("TOP", statsframe, "TOP", 0, -10)
	statsText:SetJustifyH("LEFT")
	-- Gear Frame END
	-- Equipment Frame END

	-- Achievement & Statistics Frame START
	local stats = CreateFrame("Frame", "CharScanner_MainFrame_Stats", mainframe)
	TemplateBorderRBL(stats)
	stats:SetWidth(400)
	stats:SetHeight(52)
	stats:SetPoint("TOPLEFT", equipment, "BOTTOMLEFT", 0, 0)
	stats:EnableMouse(true)
	stats:Hide()

	-- title START
	local button = CreateFrame("Button", "CharScanner_MainFrame_Stats_Title", stats)
	TemplateStats(button, false, "title")
	button:SetPoint("TOPLEFT", stats, "TOPLEFT", 5, -2)
	-- title END

	local TEMPstats = {}
	-- level 1 START
	for i = 1, #DATA do
		local button = CreateFrame("Button", "CharScanner_MainFrame_StatsLevel1_"..DATA[i].name, stats)
		TemplateStats(button, true, "level1")
		if i == 1 then
			button:SetPoint("TOPLEFT", "CharScanner_MainFrame_Stats_Title", "BOTTOMLEFT", 0, 0)
		else
			button:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel1_"..DATA[i-1].name, "BOTTOMLEFT", 0, 0)
		end
		button:SetScript("OnClick", function() CharScanner:NodeClick(1, DATA[i].name, i) end)
		--_G["CharScanner_MainFrame_StatsLevel1_"..DATA[i].name.."Name"]:SetText("StatsLevel1_"..DATA[i].name) -- TEST
		_G["CharScanner_MainFrame_StatsLevel1_"..DATA[i].name.."Name"]:SetTextColor(color.Level1.rgb[1], color.Level1.rgb[2], color.Level1.rgb[3], 1)
		_G["CharScanner_MainFrame_StatsLevel1_"..DATA[i].name.."Achievement"]:SetTextColor(color.Level1.rgb[1], color.Level1.rgb[2], color.Level1.rgb[3], 1)
		_G["CharScanner_MainFrame_StatsLevel1_"..DATA[i].name.."Statistic"]:SetTextColor(color.Level1.rgb[1], color.Level1.rgb[2], color.Level1.rgb[3], 1)
		_G["CharScanner_MainFrame_StatsLevel1_"..DATA[i].name.."BossesTotal"]:SetTextColor(color.Level1.rgb[1], color.Level1.rgb[2], color.Level1.rgb[3], 1)

		local n = "CharScanner_MainFrame_StatsLevel1_"..DATA[i].name
		TEMPstats[n] = {}
		TEMPstats[n].level = 1
		TEMPstats[n].index1 = i
		TEMPstats[n].index2 = 0
		TEMPstats[n].index3 = 0
		TEMPstats[n].height = StatsBarHeight
		TEMPstats[n].collapse = false
	end
	-- level 1 END

	-- level 2 START
	for i = 1, #DATA do
		if DATA[i].statistic then
			local height = 0
			local statsLvL2 = CreateFrame("Frame", "CharScanner_MainFrame_StatsLevel2_"..DATA[i].name, _G["CharScanner_MainFrame_StatsLevel1_"..DATA[i].name])
			statsLvL2:SetWidth(400)
			statsLvL2:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel1_"..DATA[i].name, "BOTTOMLEFT", 0, 0)
			for j = 1, #DATA[i].statistic do
				local button = CreateFrame("Button", "CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..j, statsLvL2)
				TemplateStats(button, false, "level2")
				if j == 1 then
					button:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel2_"..DATA[i].name, "TOPLEFT", 0, 0)
				else
					button:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..(j-1), "BOTTOMLEFT", 0, 0)
				end
				--_G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..j.."Name"]:SetText("StatsLevel2_"..DATA[i].name..j) -- TEST
				height = height + 1
			end
			statsLvL2:SetHeight(height*StatsBarHeight)
			statsLvL2:Hide()

			local n = "CharScanner_MainFrame_StatsLevel2_"..DATA[i].name
			TEMPstats[n] = {}
			TEMPstats[n].level = 2
			TEMPstats[n].index1 = i
			TEMPstats[n].index2 = 0
			TEMPstats[n].index3 = 0
			TEMPstats[n].height = height*StatsBarHeight
			TEMPstats[n].collapse = false
		else
			local height = 0
			local statsLvL2 = CreateFrame("Frame", "CharScanner_MainFrame_StatsLevel2_"..DATA[i].name, _G["CharScanner_MainFrame_StatsLevel1_"..DATA[i].name])
			statsLvL2:SetWidth(400)
			statsLvL2:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel1_"..DATA[i].name, "BOTTOMLEFT", 0, 0)
			for j = 1, #DATA[i] do
				local button = CreateFrame("Button", "CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..DATA[i][j].name, statsLvL2)
				TemplateStats(button, true, "level2")
				if j == 1 then
					button:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel2_"..DATA[i].name, "TOPLEFT", 0, 0)
				else
					button:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..DATA[i][j-1].name, "BOTTOMLEFT", 0, 0)
				end
				button:SetScript("OnClick", function() CharScanner:NodeClick(2, DATA[i].name, i, DATA[i][j].name, j) end)
				--_G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..DATA[i][j].name.."Name"]:SetText("StatsLevel2_"..DATA[i].name..DATA[i][j].name) -- TEST
				_G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..DATA[i][j].name.."Name"]:SetTextColor(color.Level2.rgb[1], color.Level2.rgb[2], color.Level2.rgb[3], 1)
				_G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..DATA[i][j].name.."Achievement"]:SetTextColor(color.Level2.rgb[1], color.Level2.rgb[2], color.Level2.rgb[3], 1)
				_G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..DATA[i][j].name.."Statistic"]:SetTextColor(color.Level2.rgb[1], color.Level2.rgb[2], color.Level2.rgb[3], 1)
				_G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..DATA[i][j].name.."BossesTotal"]:SetTextColor(color.Level2.rgb[1], color.Level2.rgb[2], color.Level2.rgb[3], 1)
				height = height + 1
			end
			statsLvL2:SetHeight(height*StatsBarHeight)
			statsLvL2:Hide()

			local n = "CharScanner_MainFrame_StatsLevel2_"..DATA[i].name
			TEMPstats[n] = {}
			TEMPstats[n].level = 2
			TEMPstats[n].index1 = i
			TEMPstats[n].index2 = 0
			TEMPstats[n].index3 = 0
			TEMPstats[n].height = height*StatsBarHeight
			TEMPstats[n].collapse = false
		end
	end
	-- level 2 END

	-- level 3 START
	for i = 1, #DATA do
		if not DATA[i].statistic then
			for j = 1, #DATA[i] do

				if DATA[i][j].statistic then
					local height = 0
					local statsLvL3 = CreateFrame("Frame", "CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name, _G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name])
					statsLvL3:SetWidth(400)
					statsLvL3:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..DATA[i][j].name, "BOTTOMLEFT", 0, 0)
					for k = 1, #DATA[i][j].statistic do
						local button = CreateFrame("Button", "CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k, statsLvL3)
						TemplateStats(button, false, "level3")
						if k == 1 then
							button:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name, "TOPLEFT", 0, 0)
						else
							button:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..(k-1), "BOTTOMLEFT", 0, 0)
						end
						--_G["CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k.."Name"]:SetText("StatsLevel3_"..DATA[i].name..DATA[i][j].name..k) -- TEST
						height = height + 1
					end
					statsLvL3:SetHeight(height*StatsBarHeight)
					statsLvL3:Hide()

					local n = "CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name
					TEMPstats[n] = {}
					TEMPstats[n].level = 3
					TEMPstats[n].index1 = i
					TEMPstats[n].index2 = j
					TEMPstats[n].index3 = 0
					TEMPstats[n].height = height*StatsBarHeight
					TEMPstats[n].collapse = false
				else
					local height = 0
					local statsLvL3 = CreateFrame("Frame", "CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name, _G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name])
					statsLvL3:SetWidth(400)
					statsLvL3:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..DATA[i][j].name, "BOTTOMLEFT", 0, 0)
					for k = 1, #DATA[i][j] do
						local button = CreateFrame("Button", "CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k, statsLvL3)
						TemplateStats(button, true, "level3")
						if k == 1 then
							button:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name, "TOPLEFT", 0, 0)
						else
							button:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..(k-1), "BOTTOMLEFT", 0, 0)
						end
						button:SetScript("OnClick", function() CharScanner:NodeClick(3, DATA[i].name, i, DATA[i][j].name, j, DATA[i][j][k].name, k) end)
						--_G["CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k.."Name"]:SetText("StatsLevel3_"..DATA[i].name..DATA[i][j].name..k) -- TEST
						_G["CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k.."Name"]:SetTextColor(color.Level3.rgb[1], color.Level3.rgb[2], color.Level3.rgb[3], 1)
						_G["CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k.."Achievement"]:SetTextColor(color.Level3.rgb[1], color.Level3.rgb[2], color.Level3.rgb[3], 1)
						_G["CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k.."Statistic"]:SetTextColor(color.Level3.rgb[1], color.Level3.rgb[2], color.Level3.rgb[3], 1)
						_G["CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k.."BossesTotal"]:SetTextColor(color.Level3.rgb[1], color.Level3.rgb[2], color.Level3.rgb[3], 1)
						height = height + 1
					end
					statsLvL3:SetHeight(height*StatsBarHeight)
					statsLvL3:Hide()

					local n = "CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name
					TEMPstats[n] = {}
					TEMPstats[n].level = 3
					TEMPstats[n].index1 = i
					TEMPstats[n].index2 = j
					TEMPstats[n].index3 = 0
					TEMPstats[n].height = height*StatsBarHeight
					TEMPstats[n].collapse = false
				end

			end
		end
	end
	-- level 3 END

	-- level 4 START
	for i = 1, #DATA do
		if not DATA[i].statistic then
			for j = 1, #DATA[i] do

				if not DATA[i][j].statistic then
					for k = 1, #DATA[i][j] do

						if DATA[i][j][k].statistic then
							local height = 0
							local statsLvL4 = CreateFrame("Frame", "CharScanner_MainFrame_StatsLevel4_"..DATA[i].name..DATA[i][j].name..DATA[i][j][k].name, _G["CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name])
							statsLvL4:SetWidth(400)
							statsLvL4:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k, "BOTTOMLEFT", 0, 0)
							for l = 1, #DATA[i][j][k].statistic do
								local button = CreateFrame("Button", "CharScanner_MainFrame_StatsLevel4_"..DATA[i].name..DATA[i][j].name..DATA[i][j][k].name..l, statsLvL4)
								TemplateStats(button, false, "level4")
								if l == 1 then
									button:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel4_"..DATA[i].name..DATA[i][j].name..DATA[i][j][k].name, "TOPLEFT", 0, 0)
								else
									button:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel4_"..DATA[i].name..DATA[i][j].name..DATA[i][j][k].name..(l-1), "BOTTOMLEFT", 0, 0)
								end
								--_G["CharScanner_MainFrame_StatsLevel4_"..DATA[i].name..DATA[i][j].name..DATA[i][j][k].name..l.."Name"]:SetText("StatsLevel4_"..DATA[i].name..DATA[i][j].name..DATA[i][j][k].name..l) -- TEST
								height = height + 1
							end
							statsLvL4:SetHeight(height*StatsBarHeight)
							statsLvL4:Hide()

							local n = "CharScanner_MainFrame_StatsLevel4_"..DATA[i].name..DATA[i][j].name..DATA[i][j][k].name
							TEMPstats[n] = {}
							TEMPstats[n].level = 4
							TEMPstats[n].index1 = i
							TEMPstats[n].index2 = j
							TEMPstats[n].index3 = k
							TEMPstats[n].height = height*StatsBarHeight
							TEMPstats[n].collapse = false
						end

					end
				end

			end
		end
	end
	-- level 4 END
	
	-- stats sort START
	local function SortedPairs(t)
		local ntab = {}
		for k,v in pairs(t) do
			table.insert(ntab,k)
		end
		table.sort(ntab,
			function(a,b)
				if t[a].level == t[b].level then
					if t[a].index1 == t[b].index1 then
						if t[a].index2 == t[b].index2 then
							if t[a].index3 == t[b].index3 then
								if a < b then
									return true
						  	end
							elseif t[a].index3 < t[b].index3 then
								return true
							end
						elseif t[a].index2 < t[b].index2 then
							return true
						end
					elseif t[a].index1 < t[b].index1 then
						return true
					end
				elseif t[a].level < t[b].level then
					return true
				end
			end
		)
		local i = 0
		local function _f(_s,_v)
			i = i + 1
			local k = ntab[i]
			if (k) then
				return k,t[k]
			end
		end
		return _f,nil,nil
	end
	
	local ni = 1
	for key, value in SortedPairs(TEMPstats) do
		GUIstats[ni] = {}
		GUIstats[ni].name     = key
		GUIstats[ni].level    = value.level
		GUIstats[ni].index1   = value.index1
		GUIstats[ni].index2   = value.index2
		GUIstats[ni].index3   = value.index3
		GUIstats[ni].height   = value.height
		GUIstats[ni].collapse = value.collapse
		ni = ni + 1
	end
	table.wipe(TEMPstats)
	-- stats sort START
	-- Achievement & Statistics Frame START

	-- Summary Frame START
	local summary = CreateFrame("Frame", "CharScanner_MainFrame_Summary", mainframe)
	TemplateBorderRBL(summary)
	summary:SetWidth(400)
	summary:SetHeight(28)
	summary:SetPoint("TOPLEFT", stats, "BOTTOMLEFT", 0, 0)
	summary:Hide()

	local button1 = CreateFrame("Button", "CharScanner_MainFrame_SummaryButton1", summary)
	TemplateStats(button1, false, "summary")
	button1:SetPoint("TOPLEFT", summary, "TOPLEFT", 5, -2)
	local button2 = CreateFrame("Button", "CharScanner_MainFrame_SummaryButton2", summary)
	TemplateStats(button2, false, "summary")
	button2:SetPoint("TOPLEFT", button1, "BOTTOMLEFT", 0, 0)
	-- Summary Frame END

	-- Status Frame START
	local status = CreateFrame("Frame", "CharScanner_MainFrame_Status", mainframe)
	TemplateBorderRBL(status)
	status:SetWidth(400)
	status:SetHeight(16)
	status:SetPoint("TOPLEFT", summary, "BOTTOMLEFT", 0, 0)
	status:EnableMouse(true)
	status:SetScript("OnShow", function() CharScanner:CheckStatusFramePosition() end)
	status:Hide()

	local statusText = status:CreateFontString("CharScanner_MainFrame_Status_Text", "ARTWORK", "GameFontNormalSmall")
	statusText:SetPoint("TOP", status, "TOP", 0, -2)
	statusText:SetTextColor(1, 1, 1, 1) -- TEST
	-- Status Frame END



	-- History Frame START
	local history = CreateFrame("Frame", "CharScanner_HistoryFrame", mainframe)
	TemplateBorderTRBL(history)
	history:SetWidth(220)
	history:SetHeight(GUImaxHistory*12+10)
	history:SetPoint("TOPRIGHT", base, "TOPLEFT", 1, 1)
	history:EnableMouse(true)
	history:Hide()

	-- History ScrollFrame START
	local historyScrollFrame = CreateFrame("ScrollFrame", "CharScanner_HistoryFrameScrollFrame", history, "FauxScrollFrameTemplate")
	historyScrollFrame:SetPoint("TOPLEFT", history, "TOPLEFT", 0, -5)
	historyScrollFrame:SetPoint("BOTTOMRIGHT", history, "BOTTOMRIGHT", -27, 5)
	historyScrollFrame:SetScript("OnVerticalScroll", function() CharScanner:OnVerticalScroll() end)
	-- scroll knob tweak START
	CharScanner_HistoryFrameScrollFrameScrollBarThumbTexture:SetWidth(16)
	CharScanner_HistoryFrameScrollFrameScrollBarThumbTexture:SetHeight(16)
	CharScanner_HistoryFrameScrollFrameScrollBarThumbTexture:SetTexCoord(0.25, 0.75, 0.25, 0.75) -- (8/32, 24/32, 8/32, 24/32)
	-- scroll knob tweak END
	for i = 1, GUImaxHistory do
		local button = CreateFrame("Button", "CharScanner_HistoryFrame_Button"..i, history)
		TemplateHistoryList(button, i)
		button:SetID(i)
		if i == 1 then
			button:SetPoint("TOPLEFT", "CharScanner_HistoryFrame", "TOPLEFT", 0, -5)
		else
			button:SetPoint("TOPLEFT", "CharScanner_HistoryFrame_Button"..(i-1), "BOTTOMLEFT", 0, 0)
		end
	end
	-- History ScrollFrame END

	-- History StatusFrame START
	local historyStatus = CreateFrame("Frame", "CharScanner_HistoryFrame_Status", history)
	TemplateBorderRBL(historyStatus)
	historyStatus:SetWidth(220)
	historyStatus:SetHeight(32)
	historyStatus:SetPoint("TOPLEFT", history, "BOTTOMLEFT", 0, 0)
	historyStatus:EnableMouse(true)
	local historyStatusDelete = CreateFrame("Button", "CharScanner_HistoryFrame_Status_ButtonDelete", historyStatus)
	TemplateHeadlineButton(historyStatusDelete, true)
	historyStatusDelete:SetText(L["Clear All"])
	historyStatusDelete:SetWidth( ceil(CharScanner_HistoryFrame_Status_ButtonDelete:GetTextWidth()) + 10 )
	historyStatusDelete:SetHeight(14)
	historyStatusDelete:SetPoint("TOP", historyStatus, "TOP", 0, -2)
	historyStatusDelete:SetScript("OnClick", function() CharScanner:DeleteCharacter() end)
	local historyStatusText = historyStatus:CreateFontString("CharScanner_HistoryFrame_Status_Text", "ARTWORK", "GameFontNormalSmall")
	historyStatusText:SetPoint("TOP", historyStatusDelete, "BOTTOM", 0, -2)
	historyStatusText:SetTextColor(1, 1, 1, 1) -- TEST
	-- History StatusFrame END
	-- History Frame END



	-- SCAN-Button Frame START
	local scanbutton = CreateFrame("Frame", "CharScanner_ScanButtonFrame", UIParent)
	scanbutton:EnableMouse(true)
	scanbutton:SetMovable(true)
	scanbutton:SetToplevel(true)
	scanbutton:SetClampedToScreen(true)
	scanbutton:Hide()
	scanbutton:SetWidth(CharScanner_Options.scanButtonSize)
	scanbutton:SetHeight(CharScanner_Options.scanButtonSize)
	scanbutton:SetScript("OnShow", function() CharScanner:ScanButtonPos() end)
	local scanbuttonButton = CreateFrame("Button", "CharScanner_ScanButtonFrame_Button", scanbutton)
	scanbuttonButton:SetPoint("TOPLEFT", 0, 0)
	scanbuttonButton:SetPoint("BOTTOMRIGHT", 0, 0)
	scanbuttonButton:SetScript("OnClick", function() CharScanner_MainFrame:Show() CharScanner:ScanUnit("target") end)
	local texture = scanbuttonButton:CreateTexture(nil, "ARTWORK")
	texture:SetPoint("TOPLEFT", 0, 0)
	texture:SetPoint("BOTTOMRIGHT", 0, 0)
	texture:SetTexture(IMAGE_Button)
	scanbuttonButton:SetNormalTexture(texture)

	local mover = CreateFrame("Frame", "CharScanner_ScanButtonFrame_Mover", scanbutton)
	mover:EnableMouse(true)
	mover:SetToplevel(true)
	mover:SetWidth(300)
	mover:SetHeight(80)
	mover:SetPoint("TOP", scanbutton, "TOP", 0, 32)
	local moverBGtop = mover:CreateTexture(nil, "ARTWORK")
	moverBGtop:SetPoint("TOPLEFT", mover, "TOPLEFT", 0, 0)
	moverBGtop:SetPoint("BOTTOMRIGHT", mover, "TOPRIGHT", 0, -28)
	moverBGtop:SetTexture(0, 0, 0, 0.7)
	local moverBGbottom = mover:CreateTexture(nil, "ARTWORK")
	moverBGbottom:SetPoint("TOPLEFT", mover, "TOPLEFT", 0, -84)
	moverBGbottom:SetPoint("BOTTOMRIGHT", 0, 0)
	moverBGbottom:SetTexture(0, 0, 0, 0.7)
	mover:SetScript("OnEnter", function() moverBGtop:SetTexture(0, 0, 0, 0.4) moverBGbottom:SetTexture(0, 0, 0, 0.4) end)
	mover:SetScript("OnLeave", function() moverBGtop:SetTexture(0, 0, 0, 0.7) moverBGbottom:SetTexture(0, 0, 0, 0.7) end)
	mover:SetScript("OnMouseDown", function() CharScanner_ScanButtonFrame:StartMoving() end)
	mover:SetScript("OnMouseUp", function() CharScanner_ScanButtonFrame:StopMovingOrSizing() CharScanner:SaveScanButtonPos() end)

	local scanbuttonScale = CreateFrame("Slider", "CharScanner_ScanButtonFrame_ButtonScale", mover, "OptionsSliderTemplate")
	CharScanner_ScanButtonFrame_ButtonScale:SetWidth(200)
	CharScanner_ScanButtonFrame_ButtonScale:SetHeight(16)
	CharScanner_ScanButtonFrame_ButtonScale:SetPoint("TOP", mover, "TOP", 0, -88)
	CharScanner_ScanButtonFrame_ButtonScaleLow:SetText("10px")
	CharScanner_ScanButtonFrame_ButtonScaleHigh:SetText("48px")
	CharScanner_ScanButtonFrame_ButtonScale:SetMinMaxValues(10, 48)
	CharScanner_ScanButtonFrame_ButtonScale:SetValueStep(2)
	CharScanner_ScanButtonFrame_ButtonScale:SetValue(CharScanner_Options.scanButtonSize)
	local valuetext = scanbuttonScale:CreateFontString("CharScanner_ScanButtonFrame_ButtonScale_valuetext", "OVERLAY", "GameFontNormalSmall")
	valuetext:SetPoint("TOP", scanbuttonScale, "BOTTOM", 0, 0)
	valuetext:SetJustifyH("CENTER")
	valuetext:SetText(CharScanner_Options.scanButtonSize.."x"..CharScanner_Options.scanButtonSize.."px")
	CharScanner_ScanButtonFrame_ButtonScale:SetScript("OnValueChanged", function(self, value)
		CharScanner_ScanButtonFrame_ButtonScale_valuetext:SetText(value.."x"..value.."px")
		CharScanner_Options.scanButtonSize = value
		CharScanner_ScanButtonFrame:SetWidth(value)
		CharScanner_ScanButtonFrame:SetHeight(value)
	end)

	local text = mover:CreateFontString(nil, "ARTWORK", "GameFontNormalSmall")
	text:SetPoint("TOP", scanbuttonScale, "BOTTOM", 0, -20)
	text:SetJustifyH("CENTER")
	text:SetWidth(300)
	text:SetText(L["mover_desc"])
	text:SetNonSpaceWrap(true)
	text:SetTextColor(1, 1, 1, 0.5)
	local scanbuttonLock = CreateFrame("Button", "CharScanner_ScanButtonFrame_ButtonLock", mover)
	TemplateHeadlineButton(scanbuttonLock)
	scanbuttonLock:SetText(L["LOCK"])
	scanbuttonLock:SetWidth( ceil(CharScanner_ScanButtonFrame_ButtonLock:GetTextWidth()) + 20 )
	scanbuttonLock:SetHeight(22)
	scanbuttonLock:SetPoint("TOP", text, "BOTTOM", 0, -4)
	scanbuttonLock:SetScript("OnClick", function()
		CharScanner_Options.scanButtonLock = true
		CharScanner_ScanButtonFrame_Mover:Hide()
		CharScanner_InterfaceOptions_ScanButtonLock:SetChecked(CharScanner_Options.scanButtonLock)
	end)

	mover:SetHeight( 28 + 56 + 4 +   16 + ceil(valuetext:GetStringHeight())   + 4 + ceil(text:GetStringHeight()) + 4 + ceil(scanbuttonLock:GetHeight()) + 4 )

	if CharScanner_Options.scanButtonLock then
		mover:Hide()
	end
	-- SCAN-Button Frame END
	
	
	
	-- Protection Frame START
	local ProtectionFrame = CreateFrame("Frame", "CharScanner_ProtectionFrame", mainframe)
	ProtectionFrame:SetToplevel(true)
	ProtectionFrame:SetFrameStrata("DIALOG")
	ProtectionFrame:SetWidth(400)
	ProtectionFrame:SetHeight(50)
	ProtectionFrame:SetPoint("TOPLEFT", mainframe, "TOPLEFT", 0, 0)
	ProtectionFrame:SetPoint("BOTTOMLEFT", summary, "BOTTOMLEFT", 0, -16) -- for status
	ProtectionFrame:EnableMouse(true)
	ProtectionFrame:Hide()
	local texture = ProtectionFrame:CreateTexture(nil, "BACKGROUND")
	texture:SetPoint("TOPLEFT", 0, 0)
	texture:SetPoint("BOTTOMRIGHT", 0, 0)
	texture:SetTexture(0, 0, 0, 0.85)
	local ProtectionButton = CreateFrame("Button", "CharScanner_ProtectionFrame_Button", ProtectionFrame)
	TemplateHeadlineButton(ProtectionButton)
	ProtectionButton:SetText(L["STOP SCAN"])
	ProtectionButton:SetWidth( ceil(CharScanner_ProtectionFrame_Button:GetTextWidth()) + 20 )
	ProtectionButton:SetHeight(22)
	ProtectionButton:SetPoint("TOP", ProtectionFrame, "TOP", 0, -71) -- 30+38+(28-22)
	ProtectionButton:SetScript("OnClick", function() CharScanner:StopGroupScan() end)
	local progressText = ProtectionFrame:CreateFontString("CharScanner_ProtectionFrame_ProgressText", "ARTWORK", "GameFontWhiteSmall")
	progressText:SetPoint("TOP", ProtectionButton, "BOTTOM", 0, -20)
	progressText:SetJustifyH("CENTER")
	progressText:SetWidth(100)
	progressText:SetText("-/-")
	local progressBackground = ProtectionFrame:CreateTexture(nil, "BORDER")
	progressBackground:SetPoint("TOP", progressText, "BOTTOM", 0, -4)
	progressBackground:SetWidth(300)
	progressBackground:SetHeight(12)
	progressBackground:SetTexture(color.DarkGrey.rgb[1], color.DarkGrey.rgb[2], color.DarkGrey.rgb[3], 1)
	local progressDone = ProtectionFrame:CreateTexture("CharScanner_ProtectionFrame_ProgressDone", "ARTWORK")
	progressDone:SetPoint("LEFT", progressBackground, "LEFT", 0, 0)
	progressDone:SetWidth(1)
	progressDone:SetHeight(12)
	progressDone:SetBlendMode("BLEND")
	progressDone:SetTexture(color.Green.rgb[1], color.Green.rgb[2], color.Green.rgb[3], 1)

	local ProtectionHistoryFrame = CreateFrame("Frame", "CharScanner_ProtectionHistoryFrame", history)
	ProtectionHistoryFrame:SetToplevel(true)
	ProtectionHistoryFrame:SetFrameStrata("DIALOG")
	ProtectionHistoryFrame:SetWidth(220)
	ProtectionHistoryFrame:SetHeight(50)
	ProtectionHistoryFrame:SetPoint("TOPLEFT", history, "TOPLEFT", 0, 0)
	ProtectionHistoryFrame:SetPoint("BOTTOMLEFT", historyStatus, "BOTTOMLEFT", 0, 0)
	ProtectionHistoryFrame:EnableMouse(true)
	ProtectionHistoryFrame:Hide()
	local texture = ProtectionHistoryFrame:CreateTexture(nil, "ARTWORK")
	texture:SetPoint("TOPLEFT", 0, 0)
	texture:SetPoint("BOTTOMRIGHT", 0, 0)
	texture:SetTexture(0, 0, 0, 0.85)
	-- Protection Frame END
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:NodeClick(lvl, name1, index1, name2, index2, name3, index3)
	local height = 2 + ((#DATA+1)*StatsBarHeight) + 2 -- 52 | 2 + ((3+1)*12) + 2
	if lvl == 1 then
	-- level 1 START
		if _G["CharScanner_MainFrame_StatsLevel2_"..name1]:IsVisible() then
			_G["CharScanner_MainFrame_StatsLevel1_"..name1.."Collapse"]:SetTexture("Interface\\Buttons\\UI-PlusButton-Up")

			local h = NodeDisplay("CharScanner_MainFrame_StatsLevel2_"..name1, false)
			CharScanner_MainFrame_Stats:SetHeight(height+h)

			if index1 == 1 then
				_G["CharScanner_MainFrame_StatsLevel1_"..DATA[2].name]:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel1_"..DATA[1].name, "BOTTOMLEFT", 0, 0)
			elseif index1 > 1 and index1 < #DATA then
				_G["CharScanner_MainFrame_StatsLevel1_"..DATA[index1+1].name]:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel1_"..DATA[index1].name, "BOTTOMLEFT", 0, 0)
			end
		else
			_G["CharScanner_MainFrame_StatsLevel1_"..name1.."Collapse"]:SetTexture("Interface\\Buttons\\UI-MinusButton-Up")
			
			local h = NodeDisplay("CharScanner_MainFrame_StatsLevel2_"..name1, true)
			CharScanner_MainFrame_Stats:SetHeight(height+h)

			if index1 == 1 then
				_G["CharScanner_MainFrame_StatsLevel1_"..DATA[2].name]:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel2_"..name1, "BOTTOMLEFT", 0, 0)
			elseif index1 > 1 and index1 < #DATA then
				_G["CharScanner_MainFrame_StatsLevel1_"..DATA[index1+1].name]:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel2_"..name1, "BOTTOMLEFT", 0, 0)
			end
		end
	-- level 1 END
	elseif lvl == 2 then
	-- level 2 START
		if _G["CharScanner_MainFrame_StatsLevel3_"..name1..name2]:IsVisible() then
			_G["CharScanner_MainFrame_StatsLevel2_"..name1..name2.."Collapse"]:SetTexture("Interface\\Buttons\\UI-PlusButton-Up")

			local h = NodeDisplay("CharScanner_MainFrame_StatsLevel3_"..name1..name2, false)
			CharScanner_MainFrame_Stats:SetHeight(height+h)

			if index2 == 1 then
				_G["CharScanner_MainFrame_StatsLevel2_"..DATA[index1].name..DATA[index1][2].name]:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel2_"..DATA[index1].name..DATA[index1][1].name, "BOTTOMLEFT", 0, 0)
			elseif index2 > 1 and index2 < #DATA[index1] then
				_G["CharScanner_MainFrame_StatsLevel2_"..DATA[index1].name..DATA[index1][index2+1].name]:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel2_"..DATA[index1].name..DATA[index1][index2].name, "BOTTOMLEFT", 0, 0)
			end
		else
			_G["CharScanner_MainFrame_StatsLevel2_"..name1..name2.."Collapse"]:SetTexture("Interface\\Buttons\\UI-MinusButton-Up")

			local h = NodeDisplay("CharScanner_MainFrame_StatsLevel3_"..name1..name2, true)
			CharScanner_MainFrame_Stats:SetHeight(height+h)

			if index2 == 1 then
				_G["CharScanner_MainFrame_StatsLevel2_"..DATA[index1].name..DATA[index1][2].name]:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel3_"..DATA[index1].name..name2, "BOTTOMLEFT", 0, 0)
			elseif index2 > 1 and index2 < #DATA[index1] then
				_G["CharScanner_MainFrame_StatsLevel2_"..DATA[index1].name..DATA[index1][index2+1].name]:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel3_"..DATA[index1].name..name2..#DATA[index1][index2], "BOTTOMLEFT", 0, 0)
				local c1 = #DATA[index1][index2]
				local c2 = #DATA[index1][index2][c1].statistic
				if _G["CharScanner_MainFrame_StatsLevel4_"..DATA[index1].name..DATA[index1][index2].name..DATA[index1][index2][c1].name..c2]:IsVisible() then
					_G["CharScanner_MainFrame_StatsLevel2_"..DATA[index1].name..DATA[index1][index2+1].name]:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel4_"..DATA[index1].name..DATA[index1][index2].name..DATA[index1][index2][c1].name..c2, "BOTTOMLEFT", 0, 0)
				end
			end
		end
	-- level 2 END
	elseif lvl == 3 then
	-- level 3 START
		if _G["CharScanner_MainFrame_StatsLevel4_"..name1..name2..name3]:IsVisible() then
			_G["CharScanner_MainFrame_StatsLevel3_"..name1..name2..index3.."Collapse"]:SetTexture("Interface\\Buttons\\UI-PlusButton-Up")

			local h = NodeDisplay("CharScanner_MainFrame_StatsLevel4_"..name1..name2..name3, false)
			CharScanner_MainFrame_Stats:SetHeight(height+h)

			if index3 == 1 then
				_G["CharScanner_MainFrame_StatsLevel3_"..DATA[index1].name..DATA[index1][index2].name.."2"]:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel3_"..DATA[index1].name..DATA[index1][index2].name.."1", "BOTTOMLEFT", 0, 0)
			elseif index3 > 1 and index3 < #DATA[index1][index2] then
				_G["CharScanner_MainFrame_StatsLevel3_"..DATA[index1].name..DATA[index1][index2].name..(index3+1)]:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel3_"..DATA[index1].name..DATA[index1][index2].name..index3, "BOTTOMLEFT", 0, 0)
			elseif index3 == #DATA[index1][index2] and index2 < #DATA[index1] then
					_G["CharScanner_MainFrame_StatsLevel2_"..DATA[index1].name..DATA[index1][index2+1].name]:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel3_"..DATA[index1].name..DATA[index1][index2].name..#DATA[index1][index2], "BOTTOMLEFT", 0, 0)
			end
		else
			_G["CharScanner_MainFrame_StatsLevel3_"..name1..name2..index3.."Collapse"]:SetTexture("Interface\\Buttons\\UI-MinusButton-Up")

			local h = NodeDisplay("CharScanner_MainFrame_StatsLevel4_"..name1..name2..name3, true)
			CharScanner_MainFrame_Stats:SetHeight(height+h)

			if index3 == 1 then
				_G["CharScanner_MainFrame_StatsLevel3_"..DATA[index1].name..DATA[index1][index2].name.."2"]:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel4_"..DATA[index1].name..DATA[index1][index2].name..DATA[index1][index2][1].name, "BOTTOMLEFT", 0, 0)
			elseif index3 > 1 and index3 < #DATA[index1][index2] then
				_G["CharScanner_MainFrame_StatsLevel3_"..DATA[index1].name..DATA[index1][index2].name..(index3+1)]:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel4_"..DATA[index1].name..DATA[index1][index2].name..DATA[index1][index2][index3].name..#DATA[index1][index2][index3].statistic, "BOTTOMLEFT", 0, 0)
			elseif index3 == #DATA[index1][index2] and index2 < #DATA[index1] then
				_G["CharScanner_MainFrame_StatsLevel2_"..DATA[index1].name..DATA[index1][index2+1].name]:SetPoint("TOPLEFT", "CharScanner_MainFrame_StatsLevel4_"..DATA[index1].name..DATA[index1][index2].name..DATA[index1][index2][index3].name..#DATA[index1][index2][index3].statistic, "BOTTOMLEFT", 0, 0)
			end
		end
	-- level 3 END
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:UpdateHistory()
	CharScanner:CreateGUIhistory()
	CharScanner:FillHistoryList()
	CharScanner:UpdateHistoryStatus()
end

function CharScanner:CreateGUIhistory()
	for k, v in pairs(SCAN) do -- clean
		if not v.time then
			SCAN[k] = nil
		end
	end
	table.wipe(GUIhistory)
	local x = 1
	for k, v in pairs(SCAN) do
		GUIhistory[x] = {}
		GUIhistory[x].guid = k
		GUIhistory[x].classEN = v.classEN
		GUIhistory[x].name = v.name
		GUIhistory[x].time = v.time
		x = x + 1
	end
	GUIhistoryCount = #GUIhistory
	NumScannedChars = #GUIhistory

	table.sort(GUIhistory, function(a, b)
		if a.classEN == b.classEN then
			if a.name < b.name then
				return true
			end
		elseif a.classEN < b.classEN then
			return true
		end
	end)
end

function CharScanner:OnVerticalScroll()
	CharScanner_HistoryFrameScrollFrameScrollBar:SetValue(arg1)
	CharScanner_HistoryFrameScrollFrame.offset = math_floor((arg1 / 12) + 0.5)
	CharScanner:FillHistoryList()
end

function CharScanner:FillHistoryList()
	local showtableNumSize = GUImaxHistory

	if GUIhistoryCount < showtableNumSize then
		for i=1, showtableNumSize do
			local button = _G["CharScanner_HistoryFrame_Button"..i]
			button:Hide()
			button.guid = nil
		end
		showtableNumSize = GUIhistoryCount
	end

	FauxScrollFrame_Update(CharScanner_HistoryFrameScrollFrame, GUIhistoryCount, showtableNumSize, 12)

	if GUIhistoryCount <= GUImaxHistory then
		CharScanner_HistoryFrameScrollFrame:Hide()
	else
		CharScanner_HistoryFrameScrollFrame:Show()
	end

	local j = 1
	local sliderPos = CharScanner_HistoryFrameScrollFrame.offset
	for i=sliderPos+1, sliderPos+showtableNumSize do
		if sliderPos < GUIhistoryCount then
			_G["CharScanner_HistoryFrame_Button"..j.."Icon"]:SetTexCoord(classes[GUIhistory[i].classEN].icon[1], classes[GUIhistory[i].classEN].icon[2], classes[GUIhistory[i].classEN].icon[3], classes[GUIhistory[i].classEN].icon[4])
			_G["CharScanner_HistoryFrame_Button"..j.."Name"]:SetText(GUIhistory[i].name)
			_G["CharScanner_HistoryFrame_Button"..j.."Name"]:SetTextColor(RAID_CLASS_COLORS[GUIhistory[i].classEN].r, RAID_CLASS_COLORS[GUIhistory[i].classEN].g, RAID_CLASS_COLORS[GUIhistory[i].classEN].b, 1)
			_G["CharScanner_HistoryFrame_Button"..j.."Time"]:SetText(GUIhistory[i].time)

			_G["CharScanner_HistoryFrame_Button"..j]:Show()
			_G["CharScanner_HistoryFrame_Button"..j].guid = GUIhistory[i].guid

			_G["CharScanner_HistoryFrame_Button"..j.."Arrow"]:Hide()
			if GUIselectedGUID == GUIhistory[i].guid then
				_G["CharScanner_HistoryFrame_Button"..j.."Arrow"]:Show()
			end

			j = j + 1
		end
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:PreShowSavedData(guid)
	CharScanner:ClearData()
	CharScanner:ScanFrameToggle(true)
	DataStep = 0
	CharScanner:ShowData(1, guid, false)
	CharScanner:ShowData(2, guid, false)
	CharScanner:ShowData(3, guid, false)
	CharScanner:ShowData(4, guid, false)
end

function CharScanner:DeleteCharacter(guid, id)
	if guid then -- delete one GUID
		NumScannedChars = NumScannedChars - 1
		if NumScannedChars == 0 then
			table.wipe(SCAN)
			GUIselectedGUID = nil
			CharScanner:ScanFrameToggle()
			CharScanner:ClearData()
			CharScanner:UpdateHistory()
		elseif NumScannedChars == 1 then
			SCAN[guid] = nil
			if guid == GUIselectedGUID then
				if id == 1 then
					CharScanner:PreShowSavedData(CharScanner_HistoryFrame_Button2.guid)
				else
					CharScanner:PreShowSavedData(CharScanner_HistoryFrame_Button1.guid)
				end
			end
			CharScanner:UpdateHistory()
		elseif NumScannedChars > 1 then
			SCAN[guid] = nil
			if guid == GUIselectedGUID then
				if id == 1 then
					CharScanner:PreShowSavedData(CharScanner_HistoryFrame_Button2.guid)
				else
					CharScanner:PreShowSavedData(_G["CharScanner_HistoryFrame_Button"..(id-1)].guid)
				end
			end
			CharScanner:UpdateHistory()
		end
	else -- delete all GUIDs
		table.wipe(SCAN)
		GUIselectedGUID = nil
		NumScannedChars = 0
		CharScanner:ScanFrameToggle()
		CharScanner:ClearData()
		CharScanner:UpdateHistory()
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:UpdateHistoryStatus()
	local text = string.format("%s %s", NumScannedChars, L["|4Char:Chars;"])
	if memusage then	-- FOR TEST/DEBUG ONLY
		collectgarbage("collect") -- uhh Am I evil?
		UpdateAddOnMemoryUsage()
		text = ceil(GetAddOnMemoryUsage("CharScanner")).."KB | "..text
	end
	CharScanner_HistoryFrame_Status_Text:SetText(text)
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:ScanGroup()
	-- ScanGroup()-is-only-available-if-we-are-in-group,-but-a-paranoia-check-is-good! START
	local partymember = GetNumPartyMembers()
	local raidmember  = GetNumRaidMembers()
	local inParty
	local inRaid
	if partymember > 0 then
		inRaid  = false
		inParty = true
	end
	if raidmember > 1 then -- 1, not 0. Think the only situation where a 1-member-raid can exist is at the end of a battleground.
		inRaid  = true
		inParty = false
	end
	if not inRaid and not inParty then return end
	-- ScanGroup()-is-only-available-if-we-are-in-group,-but-a-paranoia-check-is-good! END

	CharScanner_ProtectionFrame:Show()
	CharScanner_ProtectionHistoryFrame:Show()
	GroupScanTimerStart = GetTime()

	-- temp membertable START
	-- we build our own membertable. so we are independent from PARTY_MEMBERS_CHANGED/RAID_ROSTER_UPDATE changes
	-- raid roster changes between button click and end of groupscan are discarded!
	local t = {}
	local x = 1
	if inParty then
		for i = 1, partymember do
			local unit = "party"..i
			local guid = UnitGUID(unit)
			if guid then
				t[x] = {}
				t[x].guid = guid
				t[x].unit = unit
				x = x + 1
			end
		end
		-- add player to temp db
		t[x] = {}
		t[x].guid = playerGUID
		t[x].unit = "player"
	end
	if inRaid then
		for i = 1, raidmember do
			local unit = "raid"..i
			local guid = UnitGUID(unit)
			if guid then
				t[x] = {}
				t[x].guid = guid
				t[x].unit = unit
				x = x + 1
			end
		end
	end
	local numMember = #t
	-- temp membertable END

	GroupScanIsActive = true
	GroupScanNext = false
	local gsStart = GetTime()
	local currentScanNumber = 1

	local function scanNextMember()
		local guid = UnitGUID(t[currentScanNumber].unit)
		if guid and guid == t[currentScanNumber].guid then
			CharScanner_ProtectionFrame_ProgressText:SetText(currentScanNumber.."/"..numMember)
			CharScanner_ProtectionFrame_ProgressDone:SetWidth(currentScanNumber*300/numMember)
			CharScanner:ScanUnit(t[currentScanNumber].unit)
			currentScanNumber = currentScanNumber + 1
		else
			CharScanner_ProtectionFrame_ProgressText:SetText(currentScanNumber.."/"..numMember)
			CharScanner_ProtectionFrame_ProgressDone:SetWidth(currentScanNumber*300/numMember)
			currentScanNumber = currentScanNumber + 1
		end
	end

	local function gsTimeCheck()
		if inCombat then
			--print("GROUPSCAN END: inCombat") -- TEST
			CharScanner:StopGroupScan()
			return
		end
		if not inPartyOrRaid then
			--print("GROUPSCAN END: not inPartyOrRaid") -- TEST
			CharScanner:StopGroupScan()
			return
		end
		if (GetTime() - gsStart) >= CharScanner_Options.scanDuration or GroupScanNext then
			if currentScanNumber > numMember then
				--print("GROUPSCAN END: timeout/lastunit - number", currentScanNumber, numMember) -- TEST
				CharScanner:StopGroupScan()
				return
			end
			-- next scan
			gsStart = GetTime()
			GroupScanNext = false
			scanNextMember()
		end
	end

	CharScanner_MainFrame_Headline_ButtonGroupScan:SetScript("OnUpdate", gsTimeCheck)
	scanNextMember()
end

function CharScanner:StopGroupScan()
	CharScanner_MainFrame_Headline_ButtonGroupScan:SetScript("OnUpdate", nil)
	CharScanner:Unregister_INSPECT_TALENT_READY()
	CharScanner:Unregister_INSPECT_ACHIEVEMENT_READY()
	
	GroupScanNext = false
	GroupScanIsActive = false
	if GUIselectedGUID and SCAN[GUIselectedGUID] then
		CharScanner:PreShowSavedData(GUIselectedGUID)
		CharScanner:UpdateHistory()
	else
		CharScanner:ClearData()
		CharScanner:UpdateHistory()
		if NumScannedChars == 0 then
			--print("CLEANER!") -- TEST
			table.wipe(SCAN)
			CharScanner:ScanFrameToggle()
		end		
	end
	CharScanner_ProtectionFrame:Hide()
	CharScanner_ProtectionHistoryFrame:Hide()
	
	if CharScanner_Options.chatoutput then
		local time = 0
		local TimeElapsed = math_floor((GetTime()-GroupScanTimerStart) * mult4 + 0.5) / mult4
		if type(TimeElapsed) == "number" and TimeElapsed > 0 then
			time = TimeElapsed
		end
		if time == 1 then
			time = string.format("%s %s", time, SECONDS)
		else
			time = time.." "..SECONDS
		end
		Print(L["GroupScan finished!"].." |cff"..color.Green.hex.."("..time..")|r")
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:ScanUnit(unit)
	if inCombat then
		CharScanner_MainFrame_Status_Text:SetText(L["Scan is deactivated in combat!"])
		CharScanner_MainFrame_Status_Text:SetTextColor(color.Red.rgb[1], color.Red.rgb[2], color.Red.rgb[3], 1)
		CharScanner_MainFrame_Status:Show()
		GroupScanNext = true
		return
	end
	if AchievementFrame and AchievementFrame:IsVisible() then
		CharScanner_MainFrame_Status_Text:SetText(L["Please close AchievementFrame!"])
		CharScanner_MainFrame_Status_Text:SetTextColor(color.Red.rgb[1], color.Red.rgb[2], color.Red.rgb[3], 1)
		CharScanner_MainFrame_Status:Show()
		GroupScanNext = true
		return
	end
	if not UnitExists(unit) or not UnitIsPlayer(unit) then
		CharScanner_MainFrame_Status_Text:SetText(L["Target is NOT a player!"])
		CharScanner_MainFrame_Status_Text:SetTextColor(color.Red.rgb[1], color.Red.rgb[2], color.Red.rgb[3], 1)
		CharScanner_MainFrame_Status:Show()
		GroupScanNext = true
		return
	end
	if not CanInspect(unit) then
		CharScanner_MainFrame_Status_Text:SetText(L["Target is NOT in inspect range!"])
		CharScanner_MainFrame_Status_Text:SetTextColor(color.Red.rgb[1], color.Red.rgb[2], color.Red.rgb[3], 1)
		CharScanner_MainFrame_Status:Show()
		GroupScanNext = true
		return
	end

	-- --------------------------------------------
	-- Special cases START
	-- --------------------------------------------
	local faction
	if UnitFactionGroup(unit) == "Horde" then
		DATA[3][1].achievement[55] = 4297
		faction = "Horde"
	else
		DATA[3][1].achievement[55] = 4298
		faction = "Alliance"
	end
	-- --------------------------------------------
	-- Special cases END
	-- --------------------------------------------

	SingleScanTimerStart = GetTime()
	DataStep = 0

	if not GroupScanIsActive then
		CharScanner_ProtectionFrame:Show()
		CharScanner_ProtectionHistoryFrame:Show()
		CharScanner:ClearData()
	end
	CharScanner:ScanFrameToggle(true)
	CharScanner_MainFrame_Status_Text:SetText("..."..L["scan"].."...")
	CharScanner_MainFrame_Status_Text:SetTextColor(color.LightYellow.rgb[1], color.LightYellow.rgb[2], color.LightYellow.rgb[3], 1)

	if UnitIsUnit("player", unit) then
		IsUnitInpected = false -- "Character" -> it's me
		unit = "player"
	else
		IsUnitInpected = true -- "Inspect"
	end

	SCANGUID = UnitGUID(unit)
	if not SCANGUID then Print("GUID ERROR!", SCANGUID) return end -- TEST
	SCANGUID = "_"..SCANGUID -- previewGUID -> if a succesfuly scanned unit is scanned again with an uncomplete scan the original (complete) scan is untouched
	CreateScanDataByGUID(SCANGUID)

	if not GroupScanIsActive then
		SingleScanIsActive = true

		local function ssTimeCheck()
			if inCombat then
				--print("SINGLESCAN END: inCombat") -- TEST
				CharScanner:StopSingleScan(true)
				return
			end
			if (GetTime() - SingleScanTimerStart) >= CharScanner_Options.scanDuration then
				--print("SINGLESCAN END: timeout") -- TEST
				CharScanner:StopSingleScan(true)
				return
			end
		end

		CharScanner_MainFrame_Headline_ButtonScan:SetScript("OnUpdate", ssTimeCheck)
	end

	local name, realm = UnitName(unit)
	if realm then
		name = name.."-"..realm
	end
	SCAN[SCANGUID].name = name
	SCAN[SCANGUID].level = UnitLevel(unit)
	SCAN[SCANGUID].race = UnitRace(unit)
	SCAN[SCANGUID].faction = faction
	SCAN[SCANGUID].class, SCAN[SCANGUID].classEN = UnitClass(unit)
	SCAN[SCANGUID].guildName, SCAN[SCANGUID].guildRankName, SCAN[SCANGUID].guildRankIndex = GetGuildInfo(unit)

	CharScanner:ShowData(1, SCANGUID, true)

	-- achievementFunctions workaround START
	-- hack to avoid error messages because of achievementFunctions.selectedCategory
	if achievementFunctions and achievementFunctions.selectedCategory then
		bugfix = achievementFunctions.selectedCategory
		achievementFunctions.selectedCategory = -1
	end
	-- achievementFunctions workaround END
	ClearAchievementComparisonUnit()
	SetAchievementComparisonUnit(unit)
	CharScanner:RegisterEvent("INSPECT_ACHIEVEMENT_READY")

	NotifyInspect(unit)
	if IsUnitInpected then
		CharScanner:ItemLevelCalc("Inspect", unit)
	else
		CharScanner:ItemLevelCalc("Character", "player")
	end
	CharScanner:RegisterEvent("INSPECT_TALENT_READY")
end

function CharScanner:StopSingleScan(statusupdate)
	CharScanner_MainFrame_Headline_ButtonScan:SetScript("OnUpdate", nil)
	CharScanner:Unregister_INSPECT_TALENT_READY()
	CharScanner:Unregister_INSPECT_ACHIEVEMENT_READY()
	SingleScanIsActive = false
	CharScanner_ProtectionFrame:Hide()
	CharScanner_ProtectionHistoryFrame:Hide()
	if statusupdate then
		CharScanner_MainFrame_Status_Text:SetText(L["Scan incomplete!"])
		CharScanner_MainFrame_Status_Text:SetTextColor(color.Red.rgb[1], color.Red.rgb[2], color.Red.rgb[3], 1)
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:ShowData(step, guid, scan)
	DataStep = DataStep + 1

	if not GroupScanIsActive then
	
		if step == 1 then

			CharScanner_MainFrame_Headline_ButtonCollapse:Enable()
			CharScanner_MainFrame_Headline_ButtonCollapseBorderColor:SetTexture(0.4, 0.4, 0.4, 1)
	
			--CharScanner_MainFrame_BaseBorderColor:SetTexture(RAID_CLASS_COLORS[SCAN[guid].classEN].r, RAID_CLASS_COLORS[SCAN[guid].classEN].g, RAID_CLASS_COLORS[SCAN[guid].classEN].b)
			-- ClassIcon
			CharScanner_MainFrame_Base_ClassIcon:SetTexture(classimg)
			CharScanner_MainFrame_Base_ClassIcon:SetTexCoord(classes[SCAN[guid].classEN].icon[1], classes[SCAN[guid].classEN].icon[2], classes[SCAN[guid].classEN].icon[3], classes[SCAN[guid].classEN].icon[4])
			-- FactionIcon
			if SCAN[guid].faction == "Horde" then
				CharScanner_MainFrame_Base_FactionIcon:SetTexture(factionimgHorde)
			else
				CharScanner_MainFrame_Base_FactionIcon:SetTexture(factionimgAlliance)
			end
			CharScanner_MainFrame_Base_FactionIcon:SetTexCoord(0.046875, 0.609375, 0.03125, 0.59375) -- (3/64, 39/64, 2/64, 38/64)
			-- Name
			CharScanner_MainFrame_Base_Name:SetText(SCAN[guid].name)
			CharScanner_MainFrame_Base_Name:SetTextColor(RAID_CLASS_COLORS[SCAN[guid].classEN].r, RAID_CLASS_COLORS[SCAN[guid].classEN].g, RAID_CLASS_COLORS[SCAN[guid].classEN].b, 1)
			-- Guild
			local guild = ""
			if SCAN[guid].guildName then
				guild = GUILD..": "..SCAN[guid].guildName
				if SCAN[guid].guildRankIndex and SCAN[guid].guildRankIndex < 5 then
					guild = guild.." | "..RANK..": "..guildRank[SCAN[guid].guildRankIndex+1]
				else
					guild = guild.." | "..RANK..": -"
				end
			else
				guild = GUILD..": - "
			end
			CharScanner_MainFrame_Base_Guild:SetText(guild)
			CharScanner_MainFrame_Base_Guild:SetTextColor(color.Level2.rgb[1], color.Level2.rgb[2], color.Level2.rgb[3], 1)
			-- Level/Race/Class
			CharScanner_MainFrame_Base_Level:SetText(LEVEL.." "..SCAN[guid].level.." - "..SCAN[guid].race..", "..SCAN[guid].class)
			CharScanner_MainFrame_Base_Level:SetTextColor(RAID_CLASS_COLORS[SCAN[guid].classEN].r, RAID_CLASS_COLORS[SCAN[guid].classEN].g, RAID_CLASS_COLORS[SCAN[guid].classEN].b, 1)

		elseif step == 2 then

			-- Equipment
			CharScanner_MainFrame_Equipment_Title:SetText(L["Equipment"]..":")
			CharScanner_MainFrame_Equipment_Text1:SetText(SCAN[guid].itemlevelAvg)
			CharScanner_MainFrame_Equipment_Text2:SetText(SCAN[guid].itemlevelQuality)
	
			local w1 = ceil(CharScanner_MainFrame_Equipment_Title:GetStringWidth())
			local w2 = ceil(CharScanner_MainFrame_Equipment_Text2:GetStringWidth())
	
			if (w1+w2) >= FrameWidth then
				CharScanner_MainFrame_Equipment_Text2:SetWidth(FrameWidth-w1-20)
				CharScanner_MainFrame_Equipment:SetHeight(38)
			else
				CharScanner_MainFrame_Equipment_Text2:SetWidth(FrameWidth-w1-20)
				CharScanner_MainFrame_Equipment:SetHeight(28)
			end
	
			-- Gear
			for i = 1, #SCAN[guid].gear do
				for j = 1, #slotButton do
	
					if i == slotButton[j].id then
						local linkID = select(3, SCAN[guid].gear[i].iLink:find("item:(%d+)"))
						local itemTexture = GetItemIcon(linkID)
						if itemTexture then
							local iLink = SCAN[guid].gear[i].iLink
							_G["CharScanner_MainFrame_Gear_Slot"..j.."IconTexture"]:SetTexture(itemTexture)
							_G["CharScanner_MainFrame_Gear_Slot"..j.."Count"]:SetText("|cff"..itemQuality[SCAN[guid].gear[i].iRar+1].color..SCAN[guid].gear[i].iLvl.."|r")
							_G["CharScanner_MainFrame_Gear_Slot"..j.."Count"]:Show()
							_G["CharScanner_MainFrame_Gear_Slot"..j]:SetScript("OnEnter", function(self) GameTooltip:SetOwner(self, "ANCHOR_RIGHT") GameTooltip:SetHyperlink(iLink) end)
							_G["CharScanner_MainFrame_Gear_Slot"..j]:SetScript("OnLeave", function() GameTooltip:Hide() end)
						end
						break
					end
	
				end
			end
	
			-- GearStats START - TODO
			local statsText = ""
			for key, value in pairs(SCAN[guid].gearStats) do
				statsText = statsText.."|cff"..color.LightYellow.hex.._G[key]..":|r "..(math_floor(value * mult2 + 0.5) / mult2).."\n"
				--TEST code for future GearStats updates START
				--if not CharScanner_Options.statsTest then CharScanner_Options.statsTest = {} end
				--if not CharScanner_Options.statsTest[key] then
				--	CharScanner_Options.statsTest[key] = _G[key]
				--end
				--TEST code for future GearStats updates END
			end
			statsText = string_sub(statsText, 0, -2) -- last '\n'
			CharScanner_MainFrame_Equipment_StatsText:SetText(statsText)
			local h = ceil(CharScanner_MainFrame_Equipment_StatsText:GetStringHeight())
			local w = ceil(CharScanner_MainFrame_Equipment_StatsText:GetStringWidth())
			CharScanner_MainFrame_Equipment_StatsFrame:SetHeight(h+20)
			CharScanner_MainFrame_Equipment_StatsFrame:SetWidth(w+20)
			-- GearStats END - TODO

		elseif step == 3 then

			-- Talent
			local col
			CharScanner_MainFrame_Talent_Title:SetText(TALENTS..":")
			if SCAN[guid].talentGroups then
				if (SCAN[guid].level-9-SCAN[guid].talentGroup1pointsUnspent) > 0 then
					CharScanner_MainFrame_Talent_TalentGroup1Icon:SetTexture(SCAN[guid].talentGroup1icon)
					local text = SCAN[guid].talentGroup1name.." ("..SCAN[guid].talentGroup1pointsString..")"
					if SCAN[guid].talentGroup1pointsUnspent > 0 then
						if SCAN[guid].activeTalentGroup == 1 then
							col = "|cff"..color.Red.hex
						else
							col = "|cff"..color.Grey.hex
						end
						text = text.." - "..col..string.format(UNSPENT_TALENT_POINTS, SCAN[guid].talentGroup1pointsUnspent).."|r"
					end
					CharScanner_MainFrame_Talent_TalentGroup1Text:SetText(text)
				else
					local text = "("..SCAN[guid].talentGroup1pointsString..")"
					if SCAN[guid].talentGroup1pointsUnspent > 0 then
						if SCAN[guid].activeTalentGroup == 1 then
							col = "|cff"..color.Red.hex
						else
							col = "|cff"..color.Grey.hex
						end
						text = text.." - "..col..string.format(UNSPENT_TALENT_POINTS, SCAN[guid].talentGroup1pointsUnspent).."|r"
					end
					CharScanner_MainFrame_Talent_TalentGroup1Text:SetText(text)
				end
				if SCAN[guid].activeTalentGroup == 1 then
					CharScanner_MainFrame_Talent_TalentGroup1Text:SetTextColor(color.White.rgb[1], color.White.rgb[2], color.White.rgb[3], 1)
				else
					CharScanner_MainFrame_Talent_TalentGroup1Text:SetTextColor(color.Grey.rgb[1], color.Grey.rgb[2], color.Grey.rgb[3], 1)
				end
				if SCAN[guid].talentGroups > 1 then
					if (SCAN[guid].level-9-SCAN[guid].talentGroup2pointsUnspent) > 0 then
						CharScanner_MainFrame_Talent_TalentGroup2Icon:SetTexture(SCAN[guid].talentGroup2icon)
						local text = SCAN[guid].talentGroup2name.." ("..SCAN[guid].talentGroup2pointsString..")"
						if SCAN[guid].talentGroup2pointsUnspent > 0 then
							if SCAN[guid].activeTalentGroup == 2 then
								col = "|cff"..color.Red.hex
							else
								col = "|cff"..color.Grey.hex
							end
							text = text.." - "..col..string.format(UNSPENT_TALENT_POINTS, SCAN[guid].talentGroup2pointsUnspent).."|r"
						end
						CharScanner_MainFrame_Talent_TalentGroup2Text:SetText(text)
					else
						local text = "("..SCAN[guid].talentGroup2pointsString..")"
						if SCAN[guid].talentGroup2pointsUnspent > 0 then
							if SCAN[guid].activeTalentGroup == 2 then
								col = "|cff"..color.Red.hex
							else
								col = "|cff"..color.Grey.hex
							end
							text = text.." - "..col..string.format(UNSPENT_TALENT_POINTS, SCAN[guid].talentGroup2pointsUnspent).."|r"
						end
						CharScanner_MainFrame_Talent_TalentGroup2Text:SetText(text)
					end
					if SCAN[guid].activeTalentGroup == 2 then
						CharScanner_MainFrame_Talent_TalentGroup2Text:SetTextColor(color.White.rgb[1], color.White.rgb[2], color.White.rgb[3], 1)
					else
						CharScanner_MainFrame_Talent_TalentGroup2Text:SetTextColor(color.Grey.rgb[1], color.Grey.rgb[2], color.Grey.rgb[3], 1)
					end
				else
					CharScanner_MainFrame_Talent_TalentGroup2Icon:SetTexture(nil)
					CharScanner_MainFrame_Talent_TalentGroup2Text:SetText("")
				end
				
			else
				CharScanner_MainFrame_Talent_TalentGroup1Icon:SetTexture(nil)
				CharScanner_MainFrame_Talent_TalentGroup1Text:SetText(L["No Talents"])
				CharScanner_MainFrame_Talent_TalentGroup2Icon:SetTexture(nil)
				CharScanner_MainFrame_Talent_TalentGroup2Text:SetText("")
			end

			-- TalentTree
			CharScanner:SwitchTalentGroup(guid, SCAN[guid].activeTalentGroup)

		elseif step == 4 then

			local col
	
			CharScanner_MainFrame_Stats_TitleName:SetText("")
			CharScanner_MainFrame_Stats_TitleAchievement:SetText(L["Achievements"])
			CharScanner_MainFrame_Stats_TitleStatistic:SetText(L["Bosses"])
			CharScanner_MainFrame_Stats_TitleBossesTotal:SetText(L["Kills"])
	
			-- level 1 START
			for i = 1, #DATA do
				_G["CharScanner_MainFrame_StatsLevel1_"..DATA[i].name.."Name"]:SetText(L[DATA[i].name])
				if DATA[i].achievementTotal == SCAN[guid].DATA[i].achievementCount then
					col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
				end
				_G["CharScanner_MainFrame_StatsLevel1_"..DATA[i].name.."Achievement"]:SetText(col..SCAN[guid].DATA[i].achievementCount.."|r / "..DATA[i].achievementTotal)
	
				if DATA[i].statisticTotal == SCAN[guid].DATA[i].statisticBoss then
					col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
				end
				_G["CharScanner_MainFrame_StatsLevel1_"..DATA[i].name.."Statistic"]:SetText(col..SCAN[guid].DATA[i].statisticBoss.."|r / "..DATA[i].statisticTotal)
				_G["CharScanner_MainFrame_StatsLevel1_"..DATA[i].name.."BossesTotal"]:SetText(SCAN[guid].DATA[i].statisticBossAll)
			--end
			-- level 1 END
	
			-- level 2 START
			--for i = 1, #DATA do
				if DATA[i].statistic then
					for j = 1, #DATA[i].statistic do
						if SCAN[guid].DATA[i].statisticBossEach[j] > 0 then
							col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
						end
						_G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..j.."Name"]:SetText(col..StatisticIDTable[DATA[i].statistic[j]].."|r")
						_G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..j.."BossesTotal"]:SetText(SCAN[guid].DATA[i].statisticBossEach[j])
					end
				else
					for j = 1, #DATA[i] do
						_G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..DATA[i][j].name.."Name"]:SetText(L[DATA[i][j].name])
						if DATA[i][j].achievementTotal == SCAN[guid].DATA[i][j].achievementCount then
							col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
						end
						_G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..DATA[i][j].name.."Achievement"]:SetText(col..SCAN[guid].DATA[i][j].achievementCount.."|r / "..DATA[i][j].achievementTotal)
						if DATA[i][j].statisticTotal == SCAN[guid].DATA[i][j].statisticBoss then
							col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
						end
						_G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..DATA[i][j].name.."Statistic"]:SetText(col..SCAN[guid].DATA[i][j].statisticBoss.."|r / "..DATA[i][j].statisticTotal)
						_G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..DATA[i][j].name.."BossesTotal"]:SetText(SCAN[guid].DATA[i][j].statisticBossAll)
					--end
				--end
			--end
			-- level 2 END
	
			-- level 3 START
			--for i = 1, #DATA do
				--if not DATA[i].statistic then
					--for j = 1, #DATA[i] do
	
						if DATA[i][j].statistic then
							for k = 1, #DATA[i][j].statistic do
								if SCAN[guid].DATA[i][j].statisticBossEach[k] > 0 then
									col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
								end
								_G["CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k.."Name"]:SetText(col..StatisticIDTable[DATA[i][j].statistic[k]].."|r")
								_G["CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k.."BossesTotal"]:SetText(SCAN[guid].DATA[i][j].statisticBossEach[k])
							end
						else
							for k = 1, #DATA[i][j] do
								_G["CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k.."Name"]:SetText(L[DATA[i][j][k].name])
								if DATA[i][j][k].achievementTotal == SCAN[guid].DATA[i][j][k].achievementCount then
									col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
								end
								_G["CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k.."Achievement"]:SetText(col..SCAN[guid].DATA[i][j][k].achievementCount.."|r / "..DATA[i][j][k].achievementTotal)
								if DATA[i][j][k].statisticTotal == SCAN[guid].DATA[i][j][k].statisticBoss then
									col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
								end
								_G["CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k.."Statistic"]:SetText(col..SCAN[guid].DATA[i][j][k].statisticBoss.."|r / "..DATA[i][j][k].statisticTotal)
								_G["CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k.."BossesTotal"]:SetText(SCAN[guid].DATA[i][j][k].statisticBossAll)
							--end
						--end
	
					--end
				--end
			--end
			-- level 3 END
	
			-- level 4 START
			--for i = 1, #DATA do
				--if not DATA[i].statistic then
					--for j = 1, #DATA[i] do
	
						--if not DATA[i][j].statistic then
							--for k = 1, #DATA[i][j] do
	
								if DATA[i][j][k].statistic then
									for l = 1, #DATA[i][j][k].statistic do
										if SCAN[guid].DATA[i][j][k].statisticBossEach[l] > 0 then
											col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
										end
										_G["CharScanner_MainFrame_StatsLevel4_"..DATA[i].name..DATA[i][j].name..DATA[i][j][k].name..l.."Name"]:SetText(col..StatisticIDTable[DATA[i][j][k].statistic[l]].."|r")
										_G["CharScanner_MainFrame_StatsLevel4_"..DATA[i].name..DATA[i][j].name..DATA[i][j][k].name..l.."BossesTotal"]:SetText(SCAN[guid].DATA[i][j][k].statisticBossEach[l])
									end
								end
	
							end
						end
	
					end
				end
			end
			-- level 4 END
	
			-- summary
			CharScanner_MainFrame_SummaryButton1Achievement:SetText(SCAN[guid].statsAchievementCount.." / "..SCAN[guid].statsAchievementTotal)
			CharScanner_MainFrame_SummaryButton1Achievement:SetTextColor(color.Level1.rgb[1], color.Level1.rgb[2], color.Level1.rgb[3], 1)
			local achievement = math_floor((SCAN[guid].statsAchievementCount/SCAN[guid].statsAchievementTotal*100) * mult2 + 0.5) / mult2
			CharScanner_MainFrame_SummaryButton2Achievement:SetText(achievement.."%")
			CharScanner_MainFrame_SummaryButton2Achievement:SetTextColor(color.LightYellow.rgb[1], color.LightYellow.rgb[2], color.LightYellow.rgb[3], 1)
	
			CharScanner_MainFrame_SummaryButton1Statistic:SetText(SCAN[guid].statsStatisticCount.." / "..SCAN[guid].statsStatisticTotal)
			CharScanner_MainFrame_SummaryButton1Statistic:SetTextColor(color.Level1.rgb[1], color.Level1.rgb[2], color.Level1.rgb[3], 1)
			local statistic = math_floor((SCAN[guid].statsStatisticCount/SCAN[guid].statsStatisticTotal*100) * mult2 + 0.5) / mult2
			CharScanner_MainFrame_SummaryButton2Statistic:SetText(statistic.."%")
			CharScanner_MainFrame_SummaryButton2Statistic:SetTextColor(color.LightYellow.rgb[1], color.LightYellow.rgb[2], color.LightYellow.rgb[3], 1)
	
			CharScanner_MainFrame_SummaryButton1BossesTotal:SetText(SCAN[guid].statsStatisticBossKills)
			CharScanner_MainFrame_SummaryButton1BossesTotal:SetTextColor(color.Level1.rgb[1], color.Level1.rgb[2], color.Level1.rgb[3], 1)

		end

		CharScanner_ProtectionFrame_ProgressText:SetText(L["Step"].." "..DataStep.."/4")
		CharScanner_ProtectionFrame_ProgressDone:SetWidth(DataStep*300/4)
	
	end

	if DataStep == 4 then
		-- DevTools_Dump(SCAN[SCANGUID].gearStats) -- TEST

		if scan then
			-- scan complete: copy&delete preview scandata START
			local realGUID = string_sub(guid, 2)
			CreateScanDataByGUID(realGUID)
			SCAN[realGUID] = SCAN[guid] -- copy preview
			for k, v in pairs(SCAN) do
				if string_sub(k, 0, 1) == "_" then
					SCAN[k] = nil -- delete preview
				end
			end
			guid = realGUID
			-- scan complete: copy&delete preview scandata END

			-- talent group guid adjustment
			CharScanner_MainFrame_TalentTree_GroupSelection_Button1:SetScript("OnClick", function() CharScanner:SwitchTalentGroup(guid, 1) end)
			CharScanner_MainFrame_TalentTree_GroupSelection_Button2:SetScript("OnClick", function() CharScanner:SwitchTalentGroup(guid, 2) end)

			local time = 0
			local TimeElapsed = math_floor((GetTime()-SingleScanTimerStart) * mult4 + 0.5) / mult4
			if type(TimeElapsed) == "number" and TimeElapsed > 0 then
				time = TimeElapsed
			end
			if time == 1 then
				time = string.format("%s %s", time, SECONDS)
			else
				time = time.." "..SECONDS
			end
			SCAN[guid].scanTime = time
			CharScanner_MainFrame_Status_Text:SetText(time)
			CharScanner_MainFrame_Status_Text:SetTextColor(color.Green.rgb[1], color.Green.rgb[2], color.Green.rgb[3], 1)

			local hour, minute = GetGameTime()
			if hour < 10 then
				hour = "0"..hour
			end
			if minute < 10 then
				minute = "0"..minute
			end
			SCAN[guid].time = hour..":"..minute

			if CharScanner_Options.chatoutput then
				local txt= "|cff"..classes[SCAN[guid].classEN].color..SCAN[guid].name.."|r"
				txt = txt.." |cff"..color.Level2.hex..LEVEL.." "..SCAN[guid].level.." - "..SCAN[guid].race..",|r"
				txt = txt.." |cff"..classes[SCAN[guid].classEN].color..SCAN[guid].class.."|r"
				if SCAN[guid].guildName then
					txt = txt.." |cff"..color.Level2.hex.."("..GUILD..": "..SCAN[guid].guildName..")|r"
				end
				Print(txt)
			end
			GUIselectedGUID = guid
			if not GroupScanIsActive then
				CharScanner:UpdateHistory()
				CharScanner:StopSingleScan(false)
			end
		else
			GUIselectedGUID = guid
			CharScanner_MainFrame_Status_Text:SetText(L["SCANNED AT"]..": "..SCAN[guid].time.."  |cff"..color.Green.hex.."("..SCAN[guid].scanTime..")|r")
			CharScanner_MainFrame_Status_Text:SetTextColor(color.White.rgb[1], color.White.rgb[2], color.White.rgb[3], 1)
		end

		if not GroupScanIsActive then
			local selectedIndex
			for i = 1, GUImaxHistory do
				_G["CharScanner_HistoryFrame_Button"..i.."Arrow"]:Hide()
				if _G["CharScanner_HistoryFrame_Button"..i].guid == guid then
					selectedIndex = i
				end
			end
			if selectedIndex then
				_G["CharScanner_HistoryFrame_Button"..selectedIndex.."Arrow"]:Show()
			end

			CharScanner_MainFrame_Headline_ButtonHistory:Enable()
			CharScanner_MainFrame_Headline_ButtonHistoryBorderColor:SetTexture(0.4, 0.4, 0.4, 1)
		end

		GroupScanNext = true

	end

end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:ItemLevelCalc(who, unit)
	local SlotCount = 17
	local ItemLevel = 0
	local ItemCount = 0
	local minItem = 0
	local maxItem = 0

	for i = 1, slotsNum do
		local i1 = IQ_GetItemLevel(unit, _G[who..slots[i]], i, slots[i])
		if i1 >= 0 then
			ItemLevel = ItemLevel + i1
			ItemCount = ItemCount + 1
		end
		if i1 > 0 then
			if minItem == 0 then
				minItem = i1
			end
			if maxItem == 0 then
				maxItem = i1
			end
			if i1 < minItem then
				minItem = i1
			end
			if i1 > maxItem then
				maxItem = i1
			end
		end
	end

	if IQ_2HWeaponCheck(unit, _G[who.."MainHandSlot"], _G[who.."SecondaryHandSlot"]) then SlotCount = 16 end
	local ItemIQ = math_floor((ItemLevel/SlotCount) * mult2 + 0.5) / mult2
	local ItemQuality = IQ_GetItemQuality(unit, who)

	local ItemIQcount = SlotCount-ItemCount
	local ItemIQtxtcount = ""
	if ItemIQcount > 0 then
		ItemIQtxtcount = " - |cff"..color.Red.hex..string.format(L["%s %s not unsed"], ItemIQcount, L["|4Slot:Slots;"]).."|r"
	end

	SCAN[SCANGUID].itemlevelAvg = L["Avg"].." 0"
	if type(ItemIQ) == "number" and ItemIQ > 0 then
		SCAN[SCANGUID].itemlevelAvg = L["Avg"].." "..ItemIQ.." ("..L["min."].." "..minItem.." - "..L["max."].." "..maxItem..")"..ItemIQtxtcount
	end
	SCAN[SCANGUID].itemlevelQuality = ItemQuality
	--SCAN[SCANGUID].itemlevelQuality = "99x "..ITEM_QUALITY0_DESC..", 99x "..ITEM_QUALITY1_DESC..", 99x "..ITEM_QUALITY2_DESC..", 99x "..ITEM_QUALITY3_DESC..", 99x "..ITEM_QUALITY4_DESC..", 99x "..ITEM_QUALITY5_DESC..", 99x "..ITEM_QUALITY6_DESC..", 99x "..ITEM_ACCOUNTBOUND -- TEST

	CharScanner:ShowData(2, SCANGUID, true)
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:SwitchTalentGroup(guid, group)
	if SCAN[guid] and SCAN[guid].talents[group] then
		local activgroup = SCAN[guid].activeTalentGroup
		local button1name = "CharScanner_MainFrame_TalentTree_GroupSelection_Button1"
		local button2name = "CharScanner_MainFrame_TalentTree_GroupSelection_Button2"

		_G[button1name]:SetScript("OnClick", function() CharScanner:SwitchTalentGroup(guid, 1) end)
		_G[button2name]:SetScript("OnClick", function() CharScanner:SwitchTalentGroup(guid, 2) end)


		if SCAN[guid].talentGroups == 1 then
			_G[button1name]:Show()
			_G[button2name]:Hide()
		else
			_G[button1name]:Show()
			_G[button2name]:Show()
		end

		if activgroup == 1 then
			_G[button1name]:SetNormalFontObject("GameFontWhiteSmall")
			_G[button2name]:SetNormalFontObject("GameFontDisableSmall")
		else
			_G[button1name]:SetNormalFontObject("GameFontDisableSmall")
			_G[button2name]:SetNormalFontObject("GameFontWhiteSmall")
		end


		if group == 1 then
			_G[button1name]:Disable()
			if group == activgroup then
				_G[button1name]:SetDisabledFontObject("GameFontWhiteSmall")
			else
				_G[button1name]:SetDisabledFontObject("GameFontDisableSmall")
			end
			local border1 = _G[button1name.."Border"]
			border1:SetPoint("TOPLEFT"     ,1, -1)
			border1:SetPoint("BOTTOMLEFT"  ,1, 0)
			border1:SetPoint("TOPRIGHT"    ,-1, -1)
			border1:SetPoint("BOTTOMRIGHT" ,-1, 0)

			_G[button2name]:Enable()
			local border2 = _G[button2name.."Border"]
			border2:SetPoint("TOPLEFT"     ,1, -1)
			border2:SetPoint("BOTTOMLEFT"  ,0, 1)
			border2:SetPoint("TOPRIGHT"    ,-1, -1)
			border2:SetPoint("BOTTOMRIGHT" ,-1, 1)
		else
			_G[button1name]:Enable()
			local border1 = _G[button1name.."Border"]
			border1:SetPoint("TOPLEFT"     ,1, -1)
			border1:SetPoint("BOTTOMLEFT"  ,0, 1)
			border1:SetPoint("TOPRIGHT"    ,-1, -1)
			border1:SetPoint("BOTTOMRIGHT" ,-1, 1)

			_G[button2name]:Disable()
			if group == activgroup then
				_G[button2name]:SetDisabledFontObject("GameFontWhiteSmall")
			else
				_G[button2name]:SetDisabledFontObject("GameFontDisableSmall")
			end

			local border2 = _G[button2name.."Border"]
			border2:SetPoint("TOPLEFT"     ,1, -1)
			border2:SetPoint("BOTTOMLEFT"  ,1, 0)
			border2:SetPoint("TOPRIGHT"    ,-1, -1)
			border2:SetPoint("BOTTOMRIGHT" ,-1, 0)
		end

		local numColumn = 4
		local tier = 1
		local column = 1
		for tab = 1, myMAX_TALENT_TABS do

			local point
			if group == 1 then
				point = select(tab, string.match(SCAN[guid].talentGroup1pointsString, "(.+)/(.+)/(.+)"))
			else
				point = select(tab, string.match(SCAN[guid].talentGroup2pointsString, "(.+)/(.+)/(.+)"))
			end
			local bName   = _G["CharScanner_MainFrame_TalentTree"..tab.."Name"]
			local bPoints = _G["CharScanner_MainFrame_TalentTree"..tab.."Points"]

			bName:SetText(classes[SCAN[guid].classEN].talentTab[tab].tabName)
			bPoints:SetText(point)

			if group == activgroup then
				bName:SetTextColor(color.White.rgb[1], color.White.rgb[2], color.White.rgb[3], 1)
				bPoints:SetTextColor(color.White.rgb[1], color.White.rgb[2], color.White.rgb[3], 1)
			else
				bName:SetTextColor(color.Grey.rgb[1], color.Grey.rgb[2], color.Grey.rgb[3], 1)
				bPoints:SetTextColor(color.Grey.rgb[1], color.Grey.rgb[2], color.Grey.rgb[3], 1)
			end

			for talent = 1, myMAX_NUM_TALENTS do

				if talent == 1 then
					tier = 1
					column = 1
				elseif mod(talent, numColumn) == 1 then
					tier = tier + 1
					column = 1
				else
					column = column + 1
				end

				local buttonName = "CharScanner_MainFrame_TalentTree_Slot"..tab.."_"..tier.."_"..column

				if
					classes[SCAN[guid].classEN].talents[tab] and
					classes[SCAN[guid].classEN].talents[tab][tier] and
					classes[SCAN[guid].classEN].talents[tab][tier][column]
					then

					local rank      = SCAN[guid].talents[group][tab][tier][column]
					local icon      = classes[SCAN[guid].classEN].talents[tab][tier][column].icon
					local spellID   = classes[SCAN[guid].classEN].talents[tab][tier][column].id
					local spellLink = "talent:"..spellID..":"..(rank-1)

					_G[buttonName.."IconTexture"]:SetTexture(icon)
					_G[buttonName.."Rank"]:SetText(rank)
					_G[buttonName]:Show()
					_G[buttonName]:SetScript("OnEnter", function(self) GameTooltip:SetOwner(self, "ANCHOR_RIGHT") GameTooltip:SetHyperlink(spellLink) end)
					_G[buttonName]:SetScript("OnLeave", function() GameTooltip:Hide() end)
					if rank == 0 then
						SetItemButtonDesaturated(_G[buttonName], 1, 0.65, 0.65, 0.65)
						_G[buttonName]:SetAlpha(0.5)
						_G[buttonName.."IconTexture"]:SetVertexColor(0.5, 0.5, 0.5)
						_G[buttonName.."RankBorder"]:Hide()
						_G[buttonName.."Rank"]:Hide()
					else
						SetItemButtonDesaturated(_G[buttonName], nil)
						_G[buttonName]:SetAlpha(1)
						_G[buttonName.."IconTexture"]:SetVertexColor(1, 1, 1)
						_G[buttonName.."RankBorder"]:Show()
						_G[buttonName.."Rank"]:Show()
					end
				else
					_G[buttonName]:Hide()
					_G[buttonName]:SetScript("OnEnter", nil)
					_G[buttonName]:SetScript("OnLeave", nil)
				end

			end
		end
	else
		for tab = 1, myMAX_TALENT_TABS do
			_G["CharScanner_MainFrame_TalentTree"..tab.."Name"]:SetText("")
			_G["CharScanner_MainFrame_TalentTree"..tab.."Points"]:SetText("")
		end
		CharScanner_MainFrame_TalentTree_GroupSelection_Button1:Hide()
		CharScanner_MainFrame_TalentTree_GroupSelection_Button1:SetScript("OnClick", nil)
		CharScanner_MainFrame_TalentTree_GroupSelection_Button2:Hide()
		CharScanner_MainFrame_TalentTree_GroupSelection_Button2:SetScript("OnClick", nil)
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:ClearData()
	-- Base
	CharScanner_MainFrame_Base_ClassIcon:SetTexture(nil)
	CharScanner_MainFrame_Base_FactionIcon:SetTexture(nil)
	CharScanner_MainFrame_Base_Name:SetText("")
	CharScanner_MainFrame_Base_Guild:SetText("")
	CharScanner_MainFrame_Base_Level:SetText("")

	-- Equipment
	CharScanner_MainFrame_Equipment_Title:SetText("")
	CharScanner_MainFrame_Equipment_Text1:SetText("")
	CharScanner_MainFrame_Equipment_Text2:SetText("")

	-- Gear
	for i = 1, #slotButton do
		_G["CharScanner_MainFrame_Gear_Slot"..i.."IconTexture"]:SetTexture(slotButton[i].texture)
		_G["CharScanner_MainFrame_Gear_Slot"..i.."Count"]:SetText("")
		_G["CharScanner_MainFrame_Gear_Slot"..i.."Count"]:Hide()
		_G["CharScanner_MainFrame_Gear_Slot"..i]:SetScript("OnEnter", nil)
		_G["CharScanner_MainFrame_Gear_Slot"..i]:SetScript("OnLeave", nil)
	end

	-- GearStats
	CharScanner_MainFrame_Equipment_StatsText:SetText("")
	CharScanner_MainFrame_Equipment_StatsFrame:SetHeight(1)
	CharScanner_MainFrame_Equipment_StatsFrame:SetWidth(1)

	-- Talent
	CharScanner_MainFrame_Talent_Title:SetText("")
	CharScanner_MainFrame_Talent_TalentGroup1Icon:SetTexture(nil)
	CharScanner_MainFrame_Talent_TalentGroup1Text:SetText("")
	CharScanner_MainFrame_Talent_TalentGroup2Icon:SetTexture(nil)
	CharScanner_MainFrame_Talent_TalentGroup2Text:SetText("")

	-- TalentTree
	CharScanner:SwitchTalentGroup(0, 0)

	local numColumn = 4
	local tier = 1
	local column = 1
	for tab = 1, myMAX_TALENT_TABS do
		for talent = 1, myMAX_NUM_TALENTS do
			if talent == 1 then
				tier = 1
				column = 1
			elseif mod(talent, numColumn) == 1 then
				tier = tier + 1
				column = 1
			else
				column = column + 1
			end
			_G["CharScanner_MainFrame_TalentTree_Slot"..tab.."_"..tier.."_"..column]:Hide()
		end
	end

	-- Stats
	CharScanner_MainFrame_Stats_TitleName:SetText("")
	CharScanner_MainFrame_Stats_TitleAchievement:SetText("")
	CharScanner_MainFrame_Stats_TitleStatistic:SetText("")
	CharScanner_MainFrame_Stats_TitleBossesTotal:SetText("")

	-- level 1 START
	for i = 1, #DATA do
		_G["CharScanner_MainFrame_StatsLevel1_"..DATA[i].name.."Name"]:SetText("")
		_G["CharScanner_MainFrame_StatsLevel1_"..DATA[i].name.."Achievement"]:SetText("")
		_G["CharScanner_MainFrame_StatsLevel1_"..DATA[i].name.."Statistic"]:SetText("")
		_G["CharScanner_MainFrame_StatsLevel1_"..DATA[i].name.."BossesTotal"]:SetText("")
	--end
	-- level 1 END

	-- level 2 START
	--for i = 1, #DATA do
		if DATA[i].statistic then
			for j = 1, #DATA[i].statistic do
				_G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..j.."Name"]:SetText("")
				_G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..j.."BossesTotal"]:SetText("")
			end
		else
			for j = 1, #DATA[i] do
				_G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..DATA[i][j].name.."Name"]:SetText("")
				_G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..DATA[i][j].name.."Achievement"]:SetText("")
				_G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..DATA[i][j].name.."Statistic"]:SetText("")
				_G["CharScanner_MainFrame_StatsLevel2_"..DATA[i].name..DATA[i][j].name.."BossesTotal"]:SetText("")
			--end
		--end
	--end
	-- level 2 END

	-- level 3 START
	--for i = 1, #DATA do
		--if not DATA[i].statistic then
			--for j = 1, #DATA[i] do

				if DATA[i][j].statistic then
					for k = 1, #DATA[i][j].statistic do
						_G["CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k.."Name"]:SetText("")
						_G["CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k.."BossesTotal"]:SetText("")
					end
				else
					for k = 1, #DATA[i][j] do
						_G["CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k.."Name"]:SetText("")
						_G["CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k.."Achievement"]:SetText("")
						_G["CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k.."Statistic"]:SetText("")
						_G["CharScanner_MainFrame_StatsLevel3_"..DATA[i].name..DATA[i][j].name..k.."BossesTotal"]:SetText("")
					--end
				--end

			--end
		--end
	--end
	-- level 3 END

	-- level 4 START
	--for i = 1, #DATA do
		--if not DATA[i].statistic then
			--for j = 1, #DATA[i] do

				--if not DATA[i][j].statistic then
					--for k = 1, #DATA[i][j] do

						if DATA[i][j][k].statistic then
							for l = 1, #DATA[i][j][k].statistic do
								_G["CharScanner_MainFrame_StatsLevel4_"..DATA[i].name..DATA[i][j].name..DATA[i][j][k].name..l.."Name"]:SetText("")
								_G["CharScanner_MainFrame_StatsLevel4_"..DATA[i].name..DATA[i][j].name..DATA[i][j][k].name..l.."BossesTotal"]:SetText("")
							end
						end

					end
				end

			end
		end
	end
	-- level 4 END

	-- summary
	CharScanner_MainFrame_SummaryButton1Achievement:SetText("")
	CharScanner_MainFrame_SummaryButton2Achievement:SetText("")

	CharScanner_MainFrame_SummaryButton1Statistic:SetText("")
	CharScanner_MainFrame_SummaryButton2Statistic:SetText("")

	CharScanner_MainFrame_SummaryButton1BossesTotal:SetText("")
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:Unregister_INSPECT_TALENT_READY()
	CharScanner:UnregisterEvent("INSPECT_TALENT_READY")
	if not InspectFrame.unit then -- hey, we do not harm any default frame. the default inspect frame is our friend.
		ClearInspectPlayer()
	end
end

function CharScanner:Unregister_INSPECT_ACHIEVEMENT_READY()
	CharScanner:UnregisterEvent("INSPECT_ACHIEVEMENT_READY")
	-- achievementFunctions workaround START
	-- hack to avoid error messages because of achievementFunctions.selectedCategory
	if achievementFunctions and achievementFunctions.selectedCategory then
		achievementFunctions.selectedCategory = bugfix
	end
	-- achievementFunctions workaround END
	ClearAchievementComparisonUnit()
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
local function OnEvent(self, event, ...)
	if event == "PLAYER_REGEN_DISABLED" then
		inCombat = true
		CharScanner_ScanButtonFrame:Hide()
	elseif event == "PLAYER_REGEN_ENABLED" then
		inCombat = false
		if CharScanner_Options.scanButton then
			if UnitIsPlayer("target") then
				CharScanner_ScanButtonFrame:Show()
			end
		end
	elseif event == "PLAYER_TARGET_CHANGED" then
		if inCombat then return end
		if CharScanner_Options.scanButton then
			if UnitIsPlayer("target") then
				CharScanner_ScanButtonFrame:Show()
			else
				CharScanner_ScanButtonFrame:Hide()
			end
		end
	elseif event == "INSPECT_TALENT_READY" then
		if not SCANGUID then Print("INSPECT_TALENT_READY: SCANGUID ERROR!", SCANGUID) end -- TEST
		SCAN[SCANGUID].talentGroups = nil
		SCAN[SCANGUID].activeTalentGroup = nil
		SCAN[SCANGUID].talentGroup1name = nil
		SCAN[SCANGUID].talentGroup1pointsString = nil
		SCAN[SCANGUID].talentGroup2name = nil
		SCAN[SCANGUID].talentGroup2pointsString = nil

		if SCAN[SCANGUID].level > 9 then
			SCAN[SCANGUID].talentGroups = GetNumTalentGroups(IsUnitInpected, nil)
			SCAN[SCANGUID].activeTalentGroup = GetActiveTalentGroup(IsUnitInpected, nil)

			-- save talent information for each class START
			-- this data is saved to the classes table if a class with level > 9 is scanned the first time
			if not classes[SCAN[SCANGUID].classEN].talentsSaved then
				for tab = 1, myMAX_TALENT_TABS do

					local tabName, _, _, background = GetTalentTabInfo(tab, IsUnitInpected, nil)
					if not classes[SCAN[SCANGUID].classEN].talentTab then classes[SCAN[SCANGUID].classEN].talentTab = {} end
					if not classes[SCAN[SCANGUID].classEN].talentTab[tab] then
						classes[SCAN[SCANGUID].classEN].talentTab[tab] = {}
						classes[SCAN[SCANGUID].classEN].talentTab[tab].tabName = tabName
						--classes[SCAN[SCANGUID].classEN].talentTab[tab].background = background
					end

					for talent=1, myMAX_NUM_TALENTS do
						local name, iconTexture, tier, column, rank = GetTalentInfo(tab, talent, IsUnitInpected, nil, nil)
						local spellLink = GetTalentLink(tab, talent, IsUnitInpected, nil, nil, nil)

						--TEST
						--if iconTexture and not spellLink then print("CharScanner INSPECT_TALENT_READY event: NO spellLink", icontexture, spellLink) end
						--TEST

						if iconTexture and spellLink then
							local spellID = select(3, spellLink:find("talent:(%d+)"))

							if not classes[SCAN[SCANGUID].classEN].talents then classes[SCAN[SCANGUID].classEN].talents = {} end
							if not classes[SCAN[SCANGUID].classEN].talents[tab] then classes[SCAN[SCANGUID].classEN].talents[tab] = {} end
							if not classes[SCAN[SCANGUID].classEN].talents[tab][tier] then classes[SCAN[SCANGUID].classEN].talents[tab][tier] = {} end
							if not classes[SCAN[SCANGUID].classEN].talents[tab][tier][column] then classes[SCAN[SCANGUID].classEN].talents[tab][tier][column] = {} end

							classes[SCAN[SCANGUID].classEN].talents[tab][tier][column].id = spellID
							classes[SCAN[SCANGUID].classEN].talents[tab][tier][column].icon = iconTexture
						end
					end
				end
				classes[SCAN[SCANGUID].classEN].talentsSaved = true
			end
			-- save talent information for each class END

			for group = 1, SCAN[SCANGUID].talentGroups do
				if not SCAN[SCANGUID].talents[group] then SCAN[SCANGUID].talents[group] = {} end
				if group == 1 then
					SCAN[SCANGUID].talentGroup1pointsUnspent = GetUnspentTalentPoints(IsUnitInpected, nil, group)
				else
					SCAN[SCANGUID].talentGroup2pointsUnspent = GetUnspentTalentPoints(IsUnitInpected, nil, group)
				end
				local points = 0
				for tab = 1, myMAX_TALENT_TABS do
					if not SCAN[SCANGUID].talents[group][tab] then SCAN[SCANGUID].talents[group][tab] = {} end

					-- --------------------
					for talent=1, myMAX_NUM_TALENTS do
						local _, iconTexture, tier, column, rank = GetTalentInfo(tab, talent, IsUnitInpected, nil, group)
						if not SCAN[SCANGUID].talents[group][tab][tier] then SCAN[SCANGUID].talents[group][tab][tier] = {} end
						if iconTexture then
							SCAN[SCANGUID].talents[group][tab][tier][column] = rank
						end
					end
					-- --------------------

					-- --------------------
					local name, icon, pointsSpent, background, previewPointsSpent = GetTalentTabInfo(tab, IsUnitInpected, nil, group)
					if name and icon and pointsSpent then
						if tab == 1 then
							if group == 1 then
								points = pointsSpent
								SCAN[SCANGUID].talentGroup1pointsString = pointsSpent
								SCAN[SCANGUID].talentGroup1name         = name
								SCAN[SCANGUID].talentGroup1icon         = icon
							else
								points = pointsSpent
								SCAN[SCANGUID].talentGroup2pointsString = pointsSpent
								SCAN[SCANGUID].talentGroup2name         = name
								SCAN[SCANGUID].talentGroup2icon         = icon
							end
						else
							if group == 1 then
								SCAN[SCANGUID].talentGroup1pointsString = SCAN[SCANGUID].talentGroup1pointsString.."/"..pointsSpent
								if pointsSpent > points then
									points = pointsSpent
									SCAN[SCANGUID].talentGroup1name = name
									SCAN[SCANGUID].talentGroup1icon = icon
								end
							else
								SCAN[SCANGUID].talentGroup2pointsString = SCAN[SCANGUID].talentGroup2pointsString.."/"..pointsSpent
								if pointsSpent > points then
									points = pointsSpent
									SCAN[SCANGUID].talentGroup2name = name
									SCAN[SCANGUID].talentGroup2icon = icon
								end
							end
						end
					end
					-- --------------------

				end
			end
		end

		CharScanner:ShowData(3, SCANGUID, true)
		CharScanner:Unregister_INSPECT_TALENT_READY()
	elseif event == "INSPECT_ACHIEVEMENT_READY" then
		if not SCANGUID then Print("INSPECT_TALENT_READY: SCANGUID ERROR!", SCANGUID) end -- TEST
		-- achievement START
		for i = 1, #DATA do
			SCAN[SCANGUID].DATA[i] = {}
			SCAN[SCANGUID].DATA[i].achievementCount = 0
			if DATA[i].achievement then
				for key, value in pairs(DATA[i].achievement) do
					if GetAchievementComparisonInfo(value, 1) then
						SCAN[SCANGUID].DATA[i].achievementCount = SCAN[SCANGUID].DATA[i].achievementCount + 1
					end
				end
				SCAN[SCANGUID].statsAchievementTotal = SCAN[SCANGUID].statsAchievementTotal + DATA[i].achievementTotal
				SCAN[SCANGUID].statsAchievementCount = SCAN[SCANGUID].statsAchievementCount + SCAN[SCANGUID].DATA[i].achievementCount
			else

				for j = 1, #DATA[i] do
					SCAN[SCANGUID].DATA[i][j] = {}
					SCAN[SCANGUID].DATA[i][j].achievementCount = 0
					if DATA[i][j].achievement then
						for key, value in pairs(DATA[i][j].achievement) do
							if GetAchievementComparisonInfo(value, 1) then
								SCAN[SCANGUID].DATA[i].achievementCount = SCAN[SCANGUID].DATA[i].achievementCount + 1
								SCAN[SCANGUID].DATA[i][j].achievementCount = SCAN[SCANGUID].DATA[i][j].achievementCount + 1
							end
						end
						SCAN[SCANGUID].statsAchievementTotal = SCAN[SCANGUID].statsAchievementTotal + DATA[i][j].achievementTotal
						SCAN[SCANGUID].statsAchievementCount = SCAN[SCANGUID].statsAchievementCount + SCAN[SCANGUID].DATA[i][j].achievementCount
					else

						for k = 1, #DATA[i][j] do
							SCAN[SCANGUID].DATA[i][j][k] = {}
							SCAN[SCANGUID].DATA[i][j][k].achievementCount = 0
							if DATA[i][j][k].achievement then
								for key, value in pairs(DATA[i][j][k].achievement) do
									if GetAchievementComparisonInfo(value, 1) then
										SCAN[SCANGUID].DATA[i].achievementCount = SCAN[SCANGUID].DATA[i].achievementCount + 1
										SCAN[SCANGUID].DATA[i][j].achievementCount = SCAN[SCANGUID].DATA[i][j].achievementCount + 1
										SCAN[SCANGUID].DATA[i][j][k].achievementCount = SCAN[SCANGUID].DATA[i][j][k].achievementCount + 1
									end
								end
								SCAN[SCANGUID].statsAchievementTotal = SCAN[SCANGUID].statsAchievementTotal + DATA[i][j][k].achievementTotal
								SCAN[SCANGUID].statsAchievementCount = SCAN[SCANGUID].statsAchievementCount + SCAN[SCANGUID].DATA[i][j][k].achievementCount
							--else
							end
						end

					end

				end
			
			end
		end
		-- achievement END

		-- statistic START
		for i = 1, #DATA do
			SCAN[SCANGUID].DATA[i].statisticBoss = 0
			SCAN[SCANGUID].DATA[i].statisticBossAll = 0
			if DATA[i].statistic then
				SCAN[SCANGUID].DATA[i].statisticBossEach = {}
				for key, value in pairs(DATA[i].statistic) do
					local q = 0
					if IsUnitInpected then
						q = tonumber(GetComparisonStatistic(value)) or 0
					else
						q = tonumber(GetStatistic(value)) or 0
					end
					SCAN[SCANGUID].DATA[i].statisticBossEach[key] = q
					if q > 0 then
						SCAN[SCANGUID].DATA[i].statisticBoss = SCAN[SCANGUID].DATA[i].statisticBoss + 1
						SCAN[SCANGUID].DATA[i].statisticBossAll = SCAN[SCANGUID].DATA[i].statisticBossAll + q
					end
				end
				SCAN[SCANGUID].statsStatisticTotal = SCAN[SCANGUID].statsStatisticTotal + DATA[i].statisticTotal
				SCAN[SCANGUID].statsStatisticCount = SCAN[SCANGUID].statsStatisticCount + SCAN[SCANGUID].DATA[i].statisticBoss
				SCAN[SCANGUID].statsStatisticBossKills = SCAN[SCANGUID].statsStatisticBossKills + SCAN[SCANGUID].DATA[i].statisticBossAll
			else

				for j = 1, #DATA[i] do
					SCAN[SCANGUID].DATA[i][j].statisticBoss = 0
					SCAN[SCANGUID].DATA[i][j].statisticBossAll = 0
					if DATA[i][j].statistic then
						SCAN[SCANGUID].DATA[i][j].statisticBossEach = {}
						for key, value in pairs(DATA[i][j].statistic) do
							local q = 0
							if IsUnitInpected then
								q = tonumber(GetComparisonStatistic(value)) or 0
							else
								q = tonumber(GetStatistic(value)) or 0
							end
							SCAN[SCANGUID].DATA[i][j].statisticBossEach[key] = q
							if q > 0 then
								SCAN[SCANGUID].DATA[i].statisticBoss = SCAN[SCANGUID].DATA[i].statisticBoss + 1
								SCAN[SCANGUID].DATA[i].statisticBossAll = SCAN[SCANGUID].DATA[i].statisticBossAll + q
								SCAN[SCANGUID].DATA[i][j].statisticBoss = SCAN[SCANGUID].DATA[i][j].statisticBoss + 1
								SCAN[SCANGUID].DATA[i][j].statisticBossAll = SCAN[SCANGUID].DATA[i][j].statisticBossAll + q
							end
						end
						SCAN[SCANGUID].statsStatisticTotal = SCAN[SCANGUID].statsStatisticTotal + DATA[i][j].statisticTotal
						SCAN[SCANGUID].statsStatisticCount = SCAN[SCANGUID].statsStatisticCount + SCAN[SCANGUID].DATA[i][j].statisticBoss
						SCAN[SCANGUID].statsStatisticBossKills = SCAN[SCANGUID].statsStatisticBossKills + SCAN[SCANGUID].DATA[i][j].statisticBossAll
					else

						for k = 1, #DATA[i][j] do
							SCAN[SCANGUID].DATA[i][j][k].statisticBoss = 0
							SCAN[SCANGUID].DATA[i][j][k].statisticBossAll = 0
							if DATA[i][j][k].statistic then
								SCAN[SCANGUID].DATA[i][j][k].statisticBossEach = {}
								for key, value in pairs(DATA[i][j][k].statistic) do
									local q = 0
									if IsUnitInpected then
										q = tonumber(GetComparisonStatistic(value)) or 0
									else
										q = tonumber(GetStatistic(value)) or 0
									end
									SCAN[SCANGUID].DATA[i][j][k].statisticBossEach[key] = q
									if q > 0 then
										SCAN[SCANGUID].DATA[i].statisticBoss = SCAN[SCANGUID].DATA[i].statisticBoss + 1
										SCAN[SCANGUID].DATA[i].statisticBossAll = SCAN[SCANGUID].DATA[i].statisticBossAll + q
										SCAN[SCANGUID].DATA[i][j].statisticBoss = SCAN[SCANGUID].DATA[i][j].statisticBoss + 1
										SCAN[SCANGUID].DATA[i][j].statisticBossAll = SCAN[SCANGUID].DATA[i][j].statisticBossAll + q
										SCAN[SCANGUID].DATA[i][j][k].statisticBoss = SCAN[SCANGUID].DATA[i][j][k].statisticBoss + 1
										SCAN[SCANGUID].DATA[i][j][k].statisticBossAll = SCAN[SCANGUID].DATA[i][j][k].statisticBossAll + q
									end
								end
								SCAN[SCANGUID].statsStatisticTotal = SCAN[SCANGUID].statsStatisticTotal + DATA[i][j][k].statisticTotal
								SCAN[SCANGUID].statsStatisticCount = SCAN[SCANGUID].statsStatisticCount + SCAN[SCANGUID].DATA[i][j][k].statisticBoss
								SCAN[SCANGUID].statsStatisticBossKills = SCAN[SCANGUID].statsStatisticBossKills + SCAN[SCANGUID].DATA[i][j][k].statisticBossAll
							--else
							end
						end

					end

				end

			end
		end
		-- statistic END

		CharScanner:ShowData(4, SCANGUID, true)
		CharScanner:Unregister_INSPECT_ACHIEVEMENT_READY()
	elseif event == "RAID_ROSTER_UPDATE" then
		if inCombat then return end
		local raidmembers = GetNumRaidMembers()
		if rosterhack ~= raidmembers then -- avoid execute this block if raidmembers are just moved around
			if raidmembers > 1 then -- 1, not 0. Think the only situation where a 1-member-raid can exist is at the end of a battleground.
				inPartyOrRaid = true
			else
				inPartyOrRaid = false
			end	
			rosterhack = raidmembers
			CharScanner:GroupScanToggle()
		end
	elseif event == "PARTY_MEMBERS_CHANGED" then
		if inCombat then return end
		if GetNumRaidMembers() > 0 then return end -- avoid execute this block if we are in raid
		if GetNumPartyMembers() > 0 then
			inPartyOrRaid = true
		else
			inPartyOrRaid = false
		end
		CharScanner:GroupScanToggle()
	elseif event == "PLAYER_ENTERING_WORLD" then
		if GroupScanIsActive then
			CharScanner:StopGroupScan()
		end
		if SingleScanIsActive then
			CharScanner:StopSingleScan(true)
		end
	elseif event == "PLAYER_LOGIN" then
		LoadAddOn("Blizzard_InspectUI") -- we need this
		--LoadAddOn("Blizzard_DebugTools") -- FOR TEST/DEBUG ONLY - DevTools_Dump
		playerGUID = UnitGUID("player")
		CharScanner:InitOptions()
		CharScanner:CreateInterfaceOptions()
		CharScanner:LDBcheck()
		CharScanner:CreateMainFrame()
		CharScanner:InitializeFrame()
		table.insert(UISpecialFrames, "CharScanner_MainFrame")
		BuildIDTable()
		CharScanner:UnregisterEvent("PLAYER_LOGIN")
	end
end
-- ---------------------------------------------------------------------------------------------------------------------
CharScanner:RegisterEvent("PLAYER_LOGIN")
CharScanner:RegisterEvent("PLAYER_ENTERING_WORLD")
CharScanner:RegisterEvent("PLAYER_REGEN_DISABLED")
CharScanner:RegisterEvent("PLAYER_REGEN_ENABLED")
CharScanner:RegisterEvent("PLAYER_TARGET_CHANGED")
CharScanner:RegisterEvent("RAID_ROSTER_UPDATE")
CharScanner:RegisterEvent("PARTY_MEMBERS_CHANGED")
CharScanner:SetScript("OnEvent", OnEvent)